#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
# 
#   http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
# Run with:
# PYTHONPATH=.. python3 simple_state_machine2.py
#
"""
This test uses the AWS python SDK boto3 to access our local ASL Workflow Engine
https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/stepfunctions.html
https://docs.aws.amazon.com/code-samples/latest/catalog/code-catalog-python-example_code-stepfunctions.html

Its behaviour is essentially the same as simple_state_machine1.py though this
test explicitly calls create_state_machine. The start_execution call will result
in a message being sent to the asl_workflow_events queue "under the hood".

This test is the rough equivalent of running the following on the AWS CLI
with item 1 to 4 differing by the --input below.

aws stepfunctions --endpoint http://localhost:4584 start-execution --state-machine-arn $STATE_MACHINE_ARN --input '{"lambda":"Success"}' --name success-execution

# InternalErrorNotHandled lambda
aws stepfunctions --endpoint http://localhost:4584 start-execution --state-machine-arn $STATE_MACHINE_ARN --input '{"lambda":"InternalErrorNotHandled"}' --name internal-error-not-handled-execution

# InternalErrorHandled lambda
aws stepfunctions --endpoint http://localhost:4584 start-execution --state-machine-arn $STATE_MACHINE_ARN --input '{"lambda":"InternalErrorHandled"}' --name internal-error-handled-execution

# Timeout lambda
aws stepfunctions --endpoint http://localhost:4584 start-execution --state-machine-arn $STATE_MACHINE_ARN --input '{"lambda":"Timeout"}' --name timeout-execution
"""

import sys
assert sys.version_info >= (3, 0) # Bomb out if not running Python3

import boto3, time
from botocore.exceptions import ClientError

from asl_workflow_engine.logger import init_logging
from asl_workflow_engine.open_tracing_factory import create_tracer

ASL = """{
    "Comment": "Test Step Function",
    "StartAt": "StartState",
    "States": {
        "StartState": {
            "Type": "Pass",
            "Next": "ChoiceState"
        },
        "ChoiceState": {
            "Type": "Choice",
            "Choices": [
                {
                    "Variable": "$.lambda",
                    "StringEquals": "InternalErrorNotHandled",
                    "Next": "InternalErrorNotHandledLambda"
                },
                {
                    "Variable": "$.lambda",
                    "StringEquals": "InternalErrorHandled",
                    "Next": "InternalErrorHandledLambda"
                },
                {
                    "Variable": "$.lambda",
                    "StringEquals": "Success",
                    "Next": "SuccessLambda"
                },
                {
                    "Variable": "$.lambda",
                    "StringEquals": "Timeout",
                    "Next": "TimeoutLambda"
                }
            ],
            "Default": "FailState"
        },
        "FailState": {
            "Type": "Fail",
            "Error": "NoLambdaError",
            "Cause": "No Matches!"
        },
        "SuccessLambda": {
            "Type": "Task",
            "Resource": "arn:aws:rpcmessage:local::function:SuccessLambda",
            "Next": "WaitState"
        },
        "InternalErrorNotHandledLambda": {
            "Type": "Task",
            "Resource": "arn:aws:rpcmessage:local::function:InternalErrorNotHandledLambda",
            "Next": "EndState"
        },
        "InternalErrorHandledLambda": {
            "Type": "Task",
            "Resource": "arn:aws:rpcmessage:local::function:InternalErrorHandledLambda",
            "Next": "EndState"
        },
        "TimeoutLambda": {
            "Type": "Task",
            "Resource": "arn:aws:rpcmessage:local::function:TimeoutLambda",
            "Next": "EndState"
        },
        "EndState": {
            "Type": "Pass",
            "End": true
        },
        "WaitState": {
            "Type": "Wait",
            "Seconds":10,
            "Next": "EndState"
        }
    }
}"""

items = ['{"lambda":"Success"}',
         '{"lambda":"InternalErrorNotHandled"}',
         '{"lambda":"InternalErrorHandled"}',
         '{"lambda":"Timeout"}']

#items = ['{"lambda":"Success"}']
#items = ['{"lambda":"InternalErrorNotHandled"}']
#items = ['{"lambda":"InternalErrorHandled"}']
#items = ['{"lambda":"Timeout"}']

if __name__ == '__main__':
    # Initialise logger
    logger = init_logging(log_name='simple_state_machine2')

    # Initialising OpenTracing. It's important to do this before the boto3.client
    # call as create_tracer "patches" boto3 to add the OpenTracing hooks.
    create_tracer("simple_state_machine2", {"implementation": "Jaeger"})

    # Initialise the boto3 client setting the endpoint_url to our local
    # ASL Workflow Engine
    # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/core/session.html#boto3.session.Session.client
    sfn = boto3.client("stepfunctions", endpoint_url="http://localhost:4584")
    state_machine_arn = "arn:aws:states:local:0123456789:stateMachine:simple_state_machine"

    def create_state_machines():
        # Create state machine using a dummy roleArn. If it already exists an
        # exception will be thrown, we ignore that but raise other exceptions.
        try:
            response = sfn.create_state_machine(
                name="simple_state_machine", definition=ASL,
                roleArn="arn:aws:iam::0123456789:role/service-role/MyRole"
            )
        except sfn.exceptions.StateMachineAlreadyExists as e:
            #print(e.response)

            response = sfn.update_state_machine(
                stateMachineArn=state_machine_arn,
                definition=ASL
            )
        except ClientError as e:
            logger.error(e)


    create_state_machines()

    # Loop through items invoking a new state machine execution for each item
    for item in items:
        try:
            response = sfn.start_execution(
                stateMachineArn=state_machine_arn,
                #name=EXECUTION_NAME, # If not specified a UUID is assigned
                input=item
            )
        except sfn.exceptions.StateMachineDoesNotExist as e:
            logger.info(e)

            create_state_machines()
            response = sfn.start_execution(
                stateMachineArn=state_machine_arn,
                #name=EXECUTION_NAME, # If not specified a UUID is assigned
                input=item
            )
        except ClientError as e:
            logger.error(e)

    time.sleep(1)  # Give OpenTracing a chance to flush buffer before exiting




#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
# 
#   http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
# Run with:
# PYTHONPATH=.. python3 producer_consumer_asyncio_headers.py
# PYTHONPATH=.. LOG_LEVEL=DEBUG python3 producer_consumer_asyncio_headers.py
#
"""
This example application is mainly for exercising the messaging API and isn't
directly related to the ASL engine per se. It is an asyncio example that creates
a bounded queue and runs a simple producer/consumer as coroutines.
"""

import sys
assert sys.version_info >= (3, 6) # Bomb out if not running Python3.6

import asyncio, json, time

from asl_workflow_engine.logger import init_logging
from asl_workflow_engine.amqp_0_9_1_messaging_asyncio import Connection, Message
from asl_workflow_engine.messaging_exceptions import *

class Worker():

    def __init__(self, name):
        self.name = name
        # Initialise logger
        self.logger = init_logging(log_name=name)
        self.count = 0

    def handler(self, message):
        #print(message)
        #print(message.body)
        self.count += 1
        print(self.count)

        # Safer but slower alternative to connection.session(auto_ack=True)
        #if self.count % 10 == 0:  # Periodically acknowledge consumed messages
        #    message.acknowledge()

    async def run(self):
        # Connect to worker queue and process data.
        connection = Connection("amqp://localhost:5672?connection_attempts=20&retry_delay=10&heartbeat=0")
        try:
            await connection.open()
            #session = await connection.session()
            session = await connection.session(auto_ack=True)   

            """
            This illustrates a fairly "fancy" Address String to demonstrate
            AMQP Headers Exchange message routing. The extended syntax is
            creating an AMQP "node" that represents a named "subscription queue"
            and the message routing is defined by the exchange-to-queue bindings:
            "x-bindings": [{"exchange": "amq.match", "queue": "HeadersSubscriptionQueue", "key": "headers", "arguments": {"x-match": "all", "service": "news", "country": "uk"}}]
            These bindings match messages 'if service == news and country == uk'.
            The example sets these as message properties in the published message.
            """
            self.consumer = await session.consumer(
                "amq.match" + '; {"node": {"x-declare": {"auto-delete": true, "queue": "' + self.name + '", "arguments": {"x-max-length": 10, "x-overflow": "reject-publish"}}, "x-bindings": [{"exchange": "amq.match", "queue": "' + self.name + '", "key": "headers", "arguments": {"x-match": "all", "service": "news", "country": "uk"}}]}}'
            )

            self.consumer.capacity = 100; # Enable consumer prefetch
            await self.consumer.set_message_listener(self.handler)

            self.producer = await session.producer("amq.match")

            self.producer.enable_exceptions(sync=True)
            #self.producer.enable_exceptions()

            waiters = []
            for i in range(1000000):
                #print("message {}".format(self.producer.next_publish_seq_no))
                message = Message(body="hello message {}".format(
                    self.producer.next_publish_seq_no)
                )
                #message.subject = "news.uk"
                message.properties["service"] = "news"
                message.properties["country"] = "uk"
                #print(message)

                try:
                    # When producer.enable_exceptions(True) is set the send()
                    # method is made awaitable by returning a Future, resolved
                    # when the broker acks the message or exceptioned if the
                    # broker nacks. For fully "synchronous" publishing one can
                    # simply do: await self.producer.send(message)
                    # but that has a serious throughput/latency impact waiting
                    # for the publisher-confirms from the broker. The impact is
                    # especially bad for the case of durable queues.
                    # https://www.rabbitmq.com/confirms.html#publisher-confirms
                    # https://www.rabbitmq.com/confirms.html#publisher-confirms-latency
                    # To mitigate this the example below publishes in batches
                    # (optimally the batch size should be roughly the same as
                    # the session capacity). We store the awaitables in a list
                    # then periodically do an asyncio.gather of the waiters
                    waiters.append(self.producer.send(message))
                    if i % 100 == 0:
                        await asyncio.gather(*waiters)
                        waiters = []

                    # When producer.enable_exceptions(False) is set the send()
                    # method will raise an exception that contains the sequence
                    # numbers of messages that have failed to publish. If no
                    # call to enable_exceptions() is made send() just does
                    # basic_publish() with no confirmation, which could result
                    # in message loss. The asyncio.sleep() in the example is
                    # necessary to periodically yield to allow consumer delivery.
                    # If an exception is thrown either increase the sleep or
                    # catch the exception and resend the nacked messages that
                    # have been recorded in exception.undelivered
                    """
                    self.producer.send(message)
                    if i % 100 == 0:
                        await asyncio.sleep(.006)
                    """
                except SendError as e:
                    #print(waiters)
                    raise e

            await connection.start(); # Wait until connection closes.
    
        except MessagingError as e:  # ConnectionError, SessionError etc.
            self.logger.error(e)

        connection.close();

def global_exception_handler(loop, context):
    # context["message"] will always be there; but context["exception"] may not
    #msg = context.get("exception", context["message"])
    #print("********** global_exception_handler **********")
    #print(msg)
    pass  # Just swallow "exception was never retrieved"

if __name__ == '__main__':
    workers = ["HeadersSubscriptionQueue"]
    loop = asyncio.get_event_loop()
    loop.set_exception_handler(global_exception_handler)
    # Create list of Worker tasks using list comprehension
    tasks = [loop.create_task(Worker(name=w).run()) for w in workers]
    # Run the tasks concurrently
    loop.run_until_complete(asyncio.gather(*tasks))
    loop.close()
    



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
# 
#   http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
# Run with:
# PYTHONPATH=.. python3 map1.py
#
"""
This test uses the AWS python SDK boto3 to access our local ASL Workflow Engine
https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/stepfunctions.html
https://docs.aws.amazon.com/code-samples/latest/catalog/code-catalog-python-example_code-stepfunctions.html

This example is a minimal example to illustrate use of the Map State
https://states-language.net/spec.html#map-state

It is a modification of the second example (Map State Example With Parameters)
from the AWS Step Functions Developer Guide made even simpler and more
self-contained by using basic Pass States.
https://github.com/awsdocs/aws-step-functions-developer-guide/blob/master/doc_source/amazon-states-language-map-state.md#map-state-example-with-parameters
"""

import sys
assert sys.version_info >= (3, 0) # Bomb out if not running Python3

import boto3, time
from botocore.exceptions import ClientError

from asl_workflow_engine.logger import init_logging
from asl_workflow_engine.open_tracing_factory import create_tracer

ASL = """{
  "Comment": "Map Example.",
  "StartAt": "Validate-All",
  "States": {
    "Validate-All": {
      "Type": "Map",
      "InputPath": "$.detail",
      "ItemsPath": "$.shipped",
      "MaxConcurrency": 0,
      "ResultPath": "$.detail.shipped",
      "Parameters": {
        "parcel.$": "$$.Map.Item.Value",
        "courier.$": "$.delivery-partner"
      },
      "Iterator": {
        "StartAt": "Validate",
        "States": {
          "Validate": {
            "Type": "Pass",
            "Comment": "The Result from each iteration should be its *effective* input, which is a JSON node that contains both the current item data from the context object, and the courier information from the 'delivery-partner' field from the Map state input.",
            "End": true
          }
        }
      },
      "End": true
    }
  }
}"""

items = ["""
{
  "ship-date": "2016-03-14T01:59:00Z",
  "detail": {
    "delivery-partner": "UQS",
    "shipped": [
      { "prod": "R31", "dest-code": 9511, "quantity": 1344 },
      { "prod": "S39", "dest-code": 9511, "quantity": 40 },
      { "prod": "R31", "dest-code": 9833, "quantity": 12 },
      { "prod": "R40", "dest-code": 9860, "quantity": 887 },
      { "prod": "R40", "dest-code": 9511, "quantity": 1220 }
    ]
  }
}
"""]

if __name__ == '__main__':
    # Initialise logger
    logger = init_logging(log_name="map1")

    # Initialising OpenTracing. It's important to do this before the boto3.client
    # call as create_tracer "patches" boto3 to add the OpenTracing hooks.
    create_tracer("map1", {"implementation": "Jaeger"})

    # Initialise the boto3 client setting the endpoint_url to our local
    # ASL Workflow Engine
    # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/core/session.html#boto3.session.Session.client
    sfn = boto3.client("stepfunctions", endpoint_url="http://localhost:4584")
    state_machine_arn = "arn:aws:states:local:0123456789:stateMachine:map1"

    def create_state_machines():
        # Create state machine using a dummy roleArn. If it already exists an
        # exception will be thrown, we ignore that but raise other exceptions.
        try:
            response = sfn.create_state_machine(
                name="map1", definition=ASL,
                roleArn="arn:aws:iam::0123456789:role/service-role/MyRole"
            )
        except sfn.exceptions.StateMachineAlreadyExists as e:
            #print(e.response)

            response = sfn.update_state_machine(
                stateMachineArn=state_machine_arn,
                definition=ASL
            )
        except ClientError as e:
            logger.error(e)


    create_state_machines()

    # Loop through items invoking a new state machine execution for each item

    for item in items:
        try:
            response = sfn.start_execution(
                stateMachineArn=state_machine_arn,
                #name=EXECUTION_NAME, # If not specified a UUID is assigned
                input=item
            )
        except sfn.exceptions.StateMachineDoesNotExist as e:
            logger.info(e)

            create_state_machines()
            response = sfn.start_execution(
                stateMachineArn=state_machine_arn,
                #name=EXECUTION_NAME, # If not specified a UUID is assigned
                input=item
            )
        except ClientError as e:
            logger.error(e)

    time.sleep(1)  # Give OpenTracing a chance to flush buffer before exiting




#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
# 
#   http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
# Run with:
# PYTHONPATH=.. python3 iterate1.py
#
"""
This test illustrates the case of passing an array of data to a Step Function
and iterating through the array, triggering execution of a child Step Function
for each item in the array.

Note that this use case is one that would fit better into the new ASL Map state

The Map state provides most benefit when calling out to Tasks that might
take some time as iterating rather than running in parallel clearly results
in cumulative time. Map states also excel where one wishes to "join" the results
and rejoin the original state machine. When one just wishes to trigger a child
state machine for each item those benefits are marginal as each iteration only
costs the state machine launch cost and those are likely to be cumulative even
in the Map state where the implementation is likely to internally iterate
through the array and launch the execution. So for this particular use case the
main benefit of the Map state is likely to be a slight improvement to the
logic and readability - but that's all TBD.

This test uses the AWS python SDK boto3 to access our local ASL Workflow Engine
https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/stepfunctions.html
https://docs.aws.amazon.com/code-samples/latest/catalog/code-catalog-python-example_code-stepfunctions.html
"""

import sys
assert sys.version_info >= (3, 0) # Bomb out if not running Python3

import boto3, time
from botocore.exceptions import ClientError

from asl_workflow_engine.logger import init_logging
from asl_workflow_engine.open_tracing_factory import create_tracer

iterate1_ASL = """{
    "Comment": "State Machine to illustrate iterating an array and passing each array item to a child Step Function",
    "StartAt": "HasIndexZero",
    "States": {
        "HasIndexZero": {
            "Type": "Choice",
            "Choices": [
                {
                    "Variable": "$[0]",
                    "IsPresent": true,
                    "Next": "StepFunctionLauncher"
                }
            ],
            "Default": "EndState"
        },
        "StepFunctionLauncher": {
            "Comment": "Launch the child Step Function. In the Parameters section we extract the first item from the array passed as input to the state and use that as the input to the child Step Function. The ResultPath of null causes the input to be passed to the output.",
            "Type": "Task",
            "Resource": "arn:aws:states:local:0123456789:states:startExecution",
            "Parameters": {  
                "Input.$": "$[0]",
                "StateMachineArn": "arn:aws:states:local:0123456789:stateMachine:child_state_machine"
            },
            "ResultPath": null,
            "Next": "HasMultipleElements"
        },
        "HasMultipleElements": {
            "Type": "Choice",
            "Choices": [
                {
                    "Variable": "$[1:]",
                    "IsPresent": true,
                    "Next": "RemoveFirstItemFromArray"
                }
            ],
            "Default": "EndState"
        },
        "RemoveFirstItemFromArray": {
            "Type": "Pass",
            "InputPath": "$[1:]",
            "Next": "StepFunctionLauncher"
        },
        "EndState": {
            "Type": "Pass",
            "End": true,
            "Parameters": []
        }
    }
}"""

child_ASL = """{
    "Comment": "Trivial Child Step Function",
    "StartAt": "StartState",
    "States": {
        "StartState": {
            "Type": "Pass",
            "End": true
        }
    }
}"""


items = ['[{"category": "reference", "author": "Nigel Rees", "title": "Sayings of the Century", "price": 8.95}, {"category": "fiction", "author": "Evelyn Waugh", "title": "Sword of Honour", "price": 12.99}, {"category": "fiction", "author": "Herman Melville", "title": "Moby Dick", "isbn": "0-553-21311-3", "price": 8.99}, {"category": "fiction", "author": "J. R. R. Tolkien", "title": "The Lord of the Rings", "isbn": "0-395-19395-8", "price": 22.99}]']

if __name__ == '__main__':
    # Initialise logger
    logger = init_logging(log_name='iterate1')

    # Initialising OpenTracing. It's important to do this before the boto3.client
    # call as create_tracer "patches" boto3 to add the OpenTracing hooks.
    create_tracer("iterate1", {"implementation": "Jaeger"})

    # Initialise the boto3 client setting the endpoint_url to our local
    # ASL Workflow Engine
    # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/core/session.html#boto3.session.Session.client
    sfn = boto3.client("stepfunctions", endpoint_url="http://localhost:4584")
    iterate1_state_machine_arn = "arn:aws:states:local:0123456789:stateMachine:iterate1_state_machine"
    child_state_machine_arn = "arn:aws:states:local:0123456789:stateMachine:child_state_machine"

    def create_state_machines():
        # Create state machines using a dummy roleArn. If it already exists an
        # exception will be thrown, we ignore that but raise other exceptions.
        try:
            response = sfn.create_state_machine(
                name="iterate1_state_machine", definition=iterate1_ASL,
                roleArn="arn:aws:iam::0123456789:role/service-role/MyRole"
            )

            response = sfn.create_state_machine(
                name="child_state_machine", definition=child_ASL,
                roleArn="arn:aws:iam::0123456789:role/service-role/MyRole"
            )
        except sfn.exceptions.StateMachineAlreadyExists as e:
            #print(e.response)

            response = sfn.update_state_machine(
                stateMachineArn=iterate1_state_machine_arn,
                definition=iterate1_ASL
            )

            response = sfn.update_state_machine(
                stateMachineArn=child_state_machine_arn,
                definition=child_ASL
            )
        except ClientError as e:
            logger.error(e)


    create_state_machines()

    # Loop through items invoking a new state machine execution for each item
    for item in items:
        try:
            response = sfn.start_execution(
                stateMachineArn=iterate1_state_machine_arn,
                #name=EXECUTION_NAME, # If not specified a UUID is assigned
                input=item
            )

            """
            Get ARN of the execution just invoked on the iterate1_state_machine.
            Note that it's more awkward trying to get the execution ARN of
            the execution that this invokes on the callee because we let it use
            a system assigned ARN, but hey illustrating this stuff is kind of
            the point of this example.
            """
            execution_arn = response["executionArn"]
            history = sfn.get_execution_history(
                executionArn=execution_arn
            )
            #print("Execution history for Launcher state machine execution:")
            #print(execution_arn)
            #print()
            #print(history)
            #print()

            """
            List all executions for the callee state machine. Do NOT
            use the "SUCCEEDED" statusFilter, because this example includes a
            Wait state which waits for around 10s, so the child execution will
            still only be in the RUNNING state by this point.
            """
            executions = sfn.list_executions(
                stateMachineArn=child_state_machine_arn
            )
            #print(executions)

        except sfn.exceptions.StateMachineDoesNotExist as e:
            logger.info(e)

            create_state_machines()
            response = sfn.start_execution(
                stateMachineArn=iterate1_state_machine_arn,
                #name=EXECUTION_NAME, # If not specified a UUID is assigned
                input=item
            )
        except ClientError as e:
            logger.error(e)

    time.sleep(1)  # Give OpenTracing a chance to flush buffer before exiting




#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
# 
#   http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
# Run with:
# PYTHONPATH=.. python3 producer_consumer.py
# PYTHONPATH=.. LOG_LEVEL=DEBUG python3 producer_consumer.py
#
"""
This example application is mainly for exercising the messaging API and isn't
directly related to the ASL engine per se. It is a BlockingConnection example
that creates a bounded queue and runs a simple producer/consumer as threads.

Even when producer.enable_exceptions() is not called (and thus broker delivery
confirms aren't enabled) this multi-threaded example using BlockingConnection
has a much lower throughput than even the asyncio example that has
enable_exceptions(sync=True)
"""

import sys
assert sys.version_info >= (3, 0) # Bomb out if not running Python3

import json, time, threading

from asl_workflow_engine.logger import init_logging
from asl_workflow_engine.amqp_0_9_1_messaging import Connection, Message
from asl_workflow_engine.messaging_exceptions import *

class Worker(threading.Thread):

    def __init__(self, name):
        super().__init__(name=name) # Call Thread constructor
        # Initialise logger
        self.logger = init_logging(log_name=name)
        self.count = 0

    def handler(self, message):
        #print(message)
        #print(message.body)
        self.count += 1
        print(self.count)

        # Safer but slower alternative to connection.session(auto_ack=True)
        #if self.count % 10 == 0:  # Periodically acknowledge consumed messages
        #    message.acknowledge()

    def run(self):
        # Connect to worker queue and process data.
        connection = Connection("amqp://localhost:5672?connection_attempts=20&retry_delay=10&heartbeat=0")
        try:
            connection.open()
            #session = connection.session()
            session = connection.session(auto_ack=True)   
            self.consumer = session.consumer(
                self.getName() + '; {"node": {"x-declare": {"auto-delete": true, "arguments": {"x-max-length": 10, "x-overflow": "reject-publish"}}}}'
                #self.getName() + '; {"node": {"x-declare": {"auto-delete": true}}}'
            )
            self.consumer.capacity = 100; # Enable consumer prefetch
            self.consumer.set_message_listener(self.handler)

            connection.start(); # Wait until connection closes.
    
        except MessagingError as e:  # ConnectionError, SessionError etc.
            self.logger.error(e)

        connection.close();

class Publisher(threading.Thread):

    def __init__(self, name):
        super().__init__(name=name) # Call Thread constructor
        # Initialise logger
        self.logger = init_logging(log_name=name)
        self.count = 0

    def run(self):
        # Connect to worker queue and process data.
        connection = Connection("amqp://localhost:5672?connection_attempts=20&retry_delay=10&heartbeat=0")
        try:
            connection.open()
            session = connection.session()
            producer = session.producer(self.getName())

            producer.enable_exceptions(sync=True)
            #producer.enable_exceptions()

            for i in range(1000000):
                #print("message {}".format(producer.next_publish_seq_no))
                message = Message(body="hello message {}".format(
                   producer.next_publish_seq_no)
                )
                #print(message)

                try:
                    # When producer.enable_exceptions(True) is set the send()
                    # method will block until message delivery is confirmed
                    # when the broker acks the message or exceptioned if the
                    # broker nacks.
                    # Note that has a serious throughput/latency impact waiting
                    # for the publisher-confirms from the broker. The impact is
                    # especially bad for the case of durable queues.
                    # https://www.rabbitmq.com/confirms.html#publisher-confirms
                    # https://www.rabbitmq.com/confirms.html#publisher-confirms-latency
                    #
                    # When producer.enable_exceptions(False) is set the send()
                    # method will raise an exception that contains the sequence
                    # numbers of messages that have failed to publish. If no
                    # call to enable_exceptions() is made send() just does
                    # basic_publish() with no confirmation, which could result
                    # in message loss.
                    producer.send(message)                    
                except SendError as e:
                    raise e

            connection.start();
    
        except MessagingError as e:  # ConnectionError, SessionError etc.
            self.logger.error(e)

        time.sleep(1)  # Enough time to ensure all messages are sent before closing.
        connection.close();
    
if __name__ == '__main__':
    workers = ["BoundedQueue"]
    for w in workers:
        worker = Worker(name = w)
        worker.start()

    time.sleep(1)  # Enough time to ensure the consumer has created the queue before publishing.

    publisher = Publisher(name = "BoundedQueue")
    publisher.start()




#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
# 
#   http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
# Run with:
# PYTHONPATH=.. python3 test_payload_template.py
# PYTHONPATH=.. LOG_LEVEL=DEBUG python3 test_payload_template.py
#
"""
This test tests the ASL Payload Template, which relates to the Parameters and
ResultSelector fields. We use a Map state in the test because it supports both
Parameters and ResultSelector.
"""

import sys
assert sys.version_info >= (3, 0) # Bomb out if not running Python3

import unittest
import json
from datetime import datetime, timezone, timedelta
from threading import Timer
from asl_workflow_engine.logger import init_logging
from asl_workflow_engine.state_engine import StateEngine

ASL = """{
    "StartAt": "ChoiceState",
    "States": {
       "ChoiceState": {
            "Type": "Choice",
            "Choices": [
                {
                    "Variable": "$.items[0]",
                    "NumericEquals": 0,
                    "Next": "Validate-All"
                }
            ]
        },
        "Validate-All": {
          "Type": "Map",
          "InputPath": "$",
          "ItemsPath": "$.items",
          "MaxConcurrency": 3,
          "ResultPath": "$",
          "Parameters": {
            "item.$": "$$.Map.Item.Value",
            "index.$": "$$.Map.Item.Index"
          },
          "ResultSelector": {
            "state.$": "$$.State.Name",
            "items.$": "$[:].item"
          },
          "Iterator": {
            "StartAt": "Validate",
            "States": {
              "Validate": {
                "Type": "Pass",
                "Comment": "The Result from each iteration should be its *effective* input, which is a JSON node that contains the current item data and its index from the context object.",
                "End": true
              }
            }
          },
          "End": true
        }

    }
}"""

context = '{"StateMachine": {"Id": "arn:aws:states:local:0123456789:stateMachine:simple_state_machine", "Definition": ' + ASL + '}}'

"""
Create a simple EventDispatcher stub so that we can test the State Engine
without requiring the messaging fabric. Rather than publishing each new state
to a queue the stub simply calls notify on the StateEngine with the new state
information. This simplistic approach should be OK for most tests, but we should
be careful of issues due to recursion so we may need to revisit this IDC.
"""
class EventDispatcherStub(object):

    def __init__(self, state_engine, config):

        """
        Create an association with the state engine and give that a reference
        back to this event dispatcher so that it can publish events and make
        use of the set_timeout time scheduler.
        """
        self.state_engine = state_engine
        self.state_engine.event_dispatcher = self
        self.message_count = -1
        self.output_event = None # Retain last event so test can check its value

    """
    This simple threaded timeout should work OK, the real timeout is actually
    implemented using Pika's connection.call_later() which is single threaded
    and handled within Pika's event loop. That approach plays much better with
    Pika's event loop.
    """
    def set_timeout(self, callback, delay):
        t = Timer(delay/1000, callback)
        t.start()
        return t

    def dispatch(self, message):
        """
        Start at -1 and increment before call to notify as this stub, unlike
        the real EventDispatcher will recursively call dispatch as the State
        Engine calls publish, this means code after the call to notify won't
        be reached when one might expect it to.
        """
        self.message_count += 1
        # The notify method expects a JSON object not a string.
        self.state_engine.notify(json.loads(message), self.message_count)

    def acknowledge(self, id):
        pass

    def publish(self, item):
        # Convert event back to JSON string for dispatching.
        self.dispatch(json.dumps(item))
    
    def broadcast(self, subject, message, carrier_properties=None):
        self.output_event = message

"""
This stubs out the real TaskDispatcher execute_task method which requires
messaging infrastructure to run whereas this test is just a state machine test.
"""
def execute_task_stub(resource_arn, parameters, callback):
    name = resource_arn.split(":")[-1]
    result = {"reply": name + " reply"}
    callback(result)


class TestPayloadTemplate(unittest.TestCase):

    def setUp(self):
        # Initialise logger
        logger = init_logging(log_name="test_payload_template")
        config = {
            "state_engine": {
                "store_url": "ASL_store.json", 
                "execution_ttl": 500
            }
        }

        state_engine = StateEngine(config)
        # Stub out the real TaskDispatcher execute_task
        state_engine.task_dispatcher.execute_task = execute_task_stub
        self.event_dispatcher = EventDispatcherStub(state_engine, config)
    
    def test_payload_template(self):
        self.event_dispatcher.dispatch('{"data": {"items": [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12]}, "context": ' + context + '}')

        #print(self.event_dispatcher.output_event)
        output_event_detail = self.event_dispatcher.output_event["detail"]
        status = output_event_detail["status"]
        output = output_event_detail["output"]
        print(status)
        print(output)
        state_first = '{"state":"Validate-All","items":[0,1,2,3,4,5,6,7,8,9,10,11,12]}'
        items_first = '{"items":[0,1,2,3,4,5,6,7,8,9,10,11,12],"state":"Validate-All"}'

        self.assertEqual(status, "SUCCEEDED")
        self.assertEqual(output == state_first or output == items_first, True)

if __name__ == '__main__':
    unittest.main()




#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
"""
With aioprometheus if you want a Summary metric the "obvious" thing to do is to
use the Summary class. The problem with that however is that it directly uses
the quantile.Estimator, which is an implementation of the Cormode, Korn,
Muthukrishnan, and Srivastava algorithm for streaming calculation of targeted
high-percentile epsilon-approximate quantiles. That Estimator retains *all*
observations in a linked list, so what happens is that over time the insert
and query time linearly degrades and gradually degrades the performance of the
application being metricated, which is far from ideal.


The "official" prometheus Python client doesn't store or expose quantiles at all.
https://github.com/prometheus/client_python#summary
This module provides a BasicSummary class that extends aioprometheus Summary to
provide a simple sum + observation count implementation as per the "official"
prometheus Python client.


The "official" prometheus Java client however does expose quantiles, but the
approach that it uses is to create a wrapper around CKMSQuantiles that maintains
a ring buffer of CKMSQuantiles to provide quantiles over a sliding time window.
https://github.com/prometheus/client_java/blob/master/simpleclient/src/main/java/io/prometheus/client/TimeWindowQuantiles.java

This module extends aioprometheus Summary with a TimeWindowSummary that
uses a custom time window Estimator.
"""

import sys
assert sys.version_info >= (3, 0)  # Bomb out if not running Python3

import time, quantile
from aioprometheus import Summary

class TimeWindowEstimator(object):
    """
    Estimator estimates quantile values from sample streams in a time- and
    memory-efficient manner subject to allowed error constraints.

    The Estimator basically follows the same API as quantile.Estimator but
    follows the pattern of the Java Summary, maintaining a ring buffer of
    quantile.Estimators to provide quantiles over a sliding windows of time.
    """
    def __init__(self, *quantiles, max_age_seconds=600, age_buckets=5):
        """Initialize an Estimator.

        Estimator is not concurrency safe.

        Args:
            quantiles: A list of floating point doubles containing the target
                quantile value and allowed error.   [(0.5, 0.01), (0.99, 0.001)]
                are the default if none are provided, signifying that the median
                will be provided at a one percent error limit and the 99th
                percentile at the a 0.1 percent error limit.
            max_age_seconds: The duration of the time window is, i.e. how long
                observations are kept before they are discarded.
                Default is 10 minutes.
            age_buckets: The number of buckets used to implement the sliding
                time window. If your time window is 10 minutes, and you have
                age_buckets=5, buckets will be switched every 2 minutes. The
                value is a trade-off between resources (memory and cpu for
                maintaining the bucket) and how smooth the time window is moved.
                Default value is 5.
        """
        self._quantiles = quantiles
        self._invariants = [quantile._Quantile(q, e) for (q, e) in quantiles]
        self._observations = 0
        self._sum = 0
        self._ring_buffer = []

        for i in range(age_buckets):
            self._ring_buffer.append(quantile.Estimator(*quantiles))

        self._current_bucket = 0
        self._last_rotate_timestamp_millis = time.time() * 1000
        self._duration_between_rotates_millis = (max_age_seconds  * 1000) / age_buckets

    def observe(self, value):
        """
        Samples an observation's value.

        Args:
            value: A numeric value signifying the value to be sampled.
        """
        self._observations += 1
        self._sum += value
        self._rotate()
        for e in self._ring_buffer:
            e.observe(float(value))

    def query(self, rank):
        """
        Retrieves the value estimate for the requested quantile rank.

        The requested quantile rank must be registered in the estimator's
        invariants a priori!

        Args:
            rank: A floating point quantile rank along the interval [0, 1].

        Returns:
            A numeric value for the quantile estimate.
        """
        current = self._rotate()
        if current._observations:
            return current.query(rank)
        else:
            """
            For the case where the current quantile.Estimator has received no
            observations, as would be the case when query is called after a
            period of no observations and an "empty" quantile.Estimator is
            rotated to the head of the ring buffer, it is unclear what the
            'best' value to return is. The Java CKMSQuantiles get() returns NaN
            https://github.com/prometheus/client_java/blob/master/simpleclient/src/main/java/io/prometheus/client/CKMSQuantiles.java
            so we do that here for consistency, however a better approach might
            be to record a reference "tail" of the ring buffer each time an
            observation is made and query that quantile.Estimator.
            """
            return float("NaN")

    def _rotate(self):
        """
        rotate the ring buffer when the time threshold has been exceeded.
        Returns:
            The quantile.Estimator from the current time window bucket.
        """
        time_since_last_rotate_millis = time.time() * 1000 - self._last_rotate_timestamp_millis

        while time_since_last_rotate_millis > self._duration_between_rotates_millis:
            self._ring_buffer[self._current_bucket] = quantile.Estimator(*self._quantiles)

            self._current_bucket += 1
            if self._current_bucket >= len(self._ring_buffer):
                self._current_bucket = 0

            time_since_last_rotate_millis -= self._duration_between_rotates_millis
            self._last_rotate_timestamp_millis += self._duration_between_rotates_millis

        return self._ring_buffer[self._current_bucket]


class TimeWindowSummary(Summary):
    """
    Extend aioprometheus Summary to use our custom time window Estimator.
    """
    def observe(self, labels, value):
        """
        Add a single observation to the summary. It is basically the same as
        aioprometheus Summary add/observe, but delegates to our custom time
        window Estimator rather than directly using quantile.Estimator.
        """
        if type(value) not in (float, int):
            raise TypeError("Summary only works with digits (int, float)")

        try:
            e = self.get_value(labels)
        except KeyError:
            # Initialize quantile estimator
            e = TimeWindowEstimator(*self.invariants)
            self.set_value(labels, e)

        e.observe(float(value))  # type: ignore


class BasicEstimator(object):
    """
    The Estimator basically follows the same API as quantile.Estimator but
    follows the pattern of the basic "official" prometheus client Summary
    to provide a simple summary comprising sum plus observation count.
    """
    def __init__(self):
        """Initialize an Estimator.
        """
        self._invariants = []  # Deliberately empty, but required by Summary.get()
        self._observations = 0
        self._sum = 0

    def observe(self, value):
        """
        Samples an observation's value.

        Args:
            value: A numeric value signifying the value to be sampled.
        """
        self._observations += 1
        self._sum += value

class BasicSummary(Summary):
    """
    Extend aioprometheus Summary to use our basic Estimator.
    """
    def observe(self, labels, value):
        """
        Add a single observation to the summary. It is basically the same as
        aioprometheus Summary add/observe, but delegates to our custom basic
        Estimator rather than directly using quantile.Estimator.
        """
        if type(value) not in (float, int):
            raise TypeError("Summary only works with digits (int, float)")

        try:
            e = self.get_value(labels)
        except KeyError:
            e = BasicEstimator()
            self.set_value(labels, e)

        e.observe(float(value))  # type: ignore




#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
# Run with:
# PYTHONPATH=.. python3 workflow_engine.py
# PYTHONPATH=.. LOG_LEVEL=DEBUG python3 workflow_engine.py
#
# Run with cProfile enabled:
# PYTHONPATH=.. python3 -m cProfile -s tottime workflow_engine.py > prof-tottime.txt
#
"""
This is the main application entry point to the ASL Workflow Engine.
This class reads the JSON configuration file config.json and stores the config
object for the rest of the application to use, it then creates and starts a
StateEngine and EventDispatcher and a Web Server to handle AWS CLI/SDK REST API.

NOTE Due to the addition of asyncio support Python 3.6 is now required.
"""

import sys
assert sys.version_info >= (3, 6)  # Bomb out if not running Python3.6


import asyncio, json, os
import threading  # Run REST API in its own thread
from asl_workflow_engine.logger import init_logging
from asl_workflow_engine.open_tracing_factory import create_tracer
from asl_workflow_engine.state_engine import StateEngine
from asl_workflow_engine.event_dispatcher import EventDispatcher

import asl_workflow_engine.rest_api
import asl_workflow_engine.rest_api_asyncio

class WorkflowEngine(object):
    def __init__(self, configuration_file):
        """
        :param configuration_file: Path to coordinator configuration file
        :type configuration_file: str
        :raises IOError: If configuration file does not exist, or is not readable
        :raises ValueError: If configuration file does not contain valid JSON
        :raises AssertionError: If configuration file does not contain the required fields
        """
        # Initialise logger
        self.logger = init_logging(log_name="asl_workflow_engine")

        # Load the configuration file.
        try:
            with open(configuration_file, "r") as fp:
                config = json.load(fp)
            self.logger.info("Creating WorkflowEngine")
        except IOError as e:
            self.logger.error(
                "Unable to read configuration file: {}".format(configuration_file)
            )
            raise
        except ValueError as e:
            self.logger.error("Configuration file does not contain valid JSON")
            raise

        # Provide defaults for any unset config key
        config["event_queue"] = config.get("event_queue", {})
        config["notifier"] = config.get("notifier", {})
        config["state_engine"] = config.get("state_engine", {})
        config["rest_api"] = config.get("rest_api", {})
        config["tracer"] = config.get("tracer", {})
        config["metrics"] = config.get("metrics", {})

        """
        Override config values if a field is set as an environment variable.
        There is also a USE_STRUCTURED_LOGGING environment variable used by
        the logger to select between automation friendly structured logging
        or more human readable "traditional" logs.
        """
        eq = config["event_queue"]
        eq["queue_name"] = os.environ.get(
            "EVENT_QUEUE_QUEUE_NAME", eq.get("queue_name")
        )
        eq["instance_id"] = os.environ.get(
            "EVENT_QUEUE_INSTANCE_ID", eq.get("instance_id")
        )
        eq["queue_type"] = os.environ.get(
            "EVENT_QUEUE_QUEUE_TYPE", eq.get("queue_type")
        )
        eq["connection_url"] = os.environ.get(
            "EVENT_QUEUE_CONNECTION_URL", eq.get("connection_url")
        )
        eq["connection_options"] = os.environ.get(
            "EVENT_QUEUE_CONNECTION_OPTIONS", eq.get("connection_options")
        )
        eq["shared_event_consumer_capacity"] = os.environ.get(
            "EVENT_QUEUE_SHARED_EVENT_CONSUMER_CAPACITY", 
            eq.get("shared_event_consumer_capacity")
        )
        eq["instance_event_consumer_capacity"] = os.environ.get(
            "EVENT_QUEUE_INSTANCE_EVENT_CONSUMER_CAPACITY", 
            eq.get("instance_event_consumer_capacity")
        )
        eq["reply_to_consumer_capacity"] = os.environ.get(
            "EVENT_QUEUE_REPLY_TO_CONSUMER_CAPACITY",
            eq.get("reply_to_consumer_capacity")
        )

        no = config["notifier"]
        no["topic"] = os.environ.get(
            "NOTIFIER_TOPIC", no.get("topic")
        )

        se = config["state_engine"]
        se["store_url"] = os.environ.get("STATE_ENGINE_STORE_URL",
                                         se.get("store_url"))
        se["execution_ttl"] = os.environ.get("STATE_ENGINE_EXECUTION_TTL",
                                             se.get("execution_ttl", 86400))

        ra = config["rest_api"]
        ra["host"] = os.environ.get("REST_API_HOST", ra.get("host"))
        ra["port"] = int(os.environ.get("REST_API_PORT", ra.get("port")))
        ra["region"] = os.environ.get("REST_API_REGION", ra.get("region"))

        tr = config["tracer"]
        tr["implementation"] = os.environ.get("TRACER_IMPLEMENTATION", 
                                              tr.get("implementation", "None"))

        # The Jaeger specific env vars are derived from this document:
        # https://www.jaegertracing.io/docs/1.22/client-features/
        sampler = tr["config"]["sampler"]
        sampler["type"] = os.environ.get(
            "JAEGER_SAMPLER_TYPE", sampler.get("type")
        )
        sampler["param"] = os.environ.get(
            "JAEGER_SAMPLER_PARAM", sampler.get("param")
        )

        metrics = config["metrics"]
        metrics["implementation"] = os.environ.get(
            "METRICS_IMPLEMENTATION", metrics.get("implementation", "None")
        )
        metrics["namespace"] = os.environ.get(
            "METRICS_NAMESPACE", metrics.get("namespace", "")
        )

        """
        Initialise opentracing.tracer before creating the StateEngine,
        EventDispatcher and RestAPIinstances.

        Call asyncio.get_event_loop() here, because if we are using asyncio we
        want the tracer to use the main asyncio event loop rather than create
        a new ThreadLoop, which is the default behaviour unless a tornado IOLoop
        is passed. In recent versions of Tornado that delegates to asyncio loop.
        """
        if eq["queue_type"].endswith("-asyncio"):
            # Attempt to use uvloop libuv based event loop if available
            # https://github.com/MagicStack/uvloop
            try:
                import uvloop
                uvloop.install()
                self.logger.info("Using uvloop asyncio event loop")
            except:  # Fall back to standard library asyncio epoll event loop
                self.logger.info("Using standard library asyncio event loop")

            loop = asyncio.get_event_loop()
            create_tracer("asl_workflow_engine", config["tracer"], use_asyncio=True)
        else:
            create_tracer("asl_workflow_engine", config["tracer"])

        self.state_engine = StateEngine(config)
        self.event_dispatcher = EventDispatcher(self.state_engine, config)

        self.config = config

    def start(self):
        if self.event_dispatcher.name.endswith("_asyncio"):
            def global_exception_handler(loop, context):
                """
                Just swallow "exception was never retrieved" as we handle the
                main exceptions that we care about in EventDispatcher.start_asyncio()
                """
                # context["message"] will always be there; but context["exception"] may not
                self.logger.error(context.get("message"))
                exception = context.get("exception")
                if exception:
                    self.logger.error(repr(exception))


            self.rest_api = asl_workflow_engine.rest_api_asyncio.RestAPI(
                self.state_engine, self.event_dispatcher, self.config
            )
            app = self.rest_api.create_app()

            loop = asyncio.get_event_loop()
            loop.set_exception_handler(global_exception_handler)
            loop.create_task(self.event_dispatcher.start_asyncio())
            """
            Start moving towards having Quart run via an external ASGI server
            so it's easier to compare the performance of different ones.
            Doing that is a little bit fiddly as at requires some refactoring
            to make the REST API the main application entry point rather than
            the WorkflowEngine class. Another complication is ensuring the
            EventDispatcher gets passed the correct event loop when refactoring.
            For now just start via the API serve function as described here:
            https://pgjones.gitlab.io/hypercorn/how_to_guides/api_usage.html
            instead of using app.run()
            """
            
            # hypercorn
            #app.run(host=self.rest_api.host, port=self.rest_api.port, loop=loop)

            from hypercorn.asyncio import serve
            from hypercorn.config import Config
            from hypercorn.run import run

            config = Config()
            config.bind = ["{}:{}".format(self.rest_api.host, self.rest_api.port)]

            loop.run_until_complete(serve(app, config))
            
            """
            # uvicorn
            import uvicorn

            # Setting loop="none" in uvicorn.run actually means use current event loop

            #import yappi
            #yappi.set_clock_type("WALL")
            #with yappi.run():

            uvicorn.run(
                app, host=self.rest_api.host, port=self.rest_api.port,
                loop="none", log_level="error"
            )

            #yappi.get_func_stats().print_all(columns={
            #    0: ("name", 140),
            #    1: ("ncall", 8),
            #    2: ("tsub", 8),
            #    3: ("ttot", 8),
            #    4: ("tavg", 8)
            #})
            """
        else:
            self.rest_api = asl_workflow_engine.rest_api.RestAPI(
                self.state_engine, self.event_dispatcher, self.config
            )
            app = self.rest_api.create_app()
            # https://stackoverflow.com/questions/31264826/start-a-flask-application-in-separate-thread/31265602#31265602
            threading.Thread(
                target=app.run,
                kwargs={
                    "host": self.rest_api.host,
                    "port": self.rest_api.port,
                },
                daemon=True,
            ).start()
            self.event_dispatcher.start()

if __name__ == "__main__":
    WorkflowEngine("config.json").start()




#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

"""
-------------------------------- READ ME FIRST ---------------------------------
Note that in JSON and dict manipulation herein there may be a mix of camel case
and fields starting with capitals. This is unfortunate, but somewhat deliberate
as we are trying to follow the patterns used in real AWS Step Functions, which
seems to use camel case in the REST API calls but in the Context object and
indeed in the ASL specification the fields start with capitals. Be aware of this
if suddenly overcome by the urge to "make everything consistent"
--------------------------------------------------------------------------------
"""

import sys
assert sys.version_info >= (3, 0)  # Bomb out if not running Python3


import operator, time, uuid, fnmatch, opentracing

from datetime import datetime, timezone, timedelta
from aioprometheus import Counter, Histogram

#from asl_workflow_engine.metrics_summary import TimeWindowSummary as Summary
from asl_workflow_engine.metrics_summary import BasicSummary as Summary

from asl_workflow_engine.state_engine_paths import (
    apply_path,
    apply_jsonpath,
    get_full_jsonpath,
    apply_resultpath,
    evaluate_payload_template,
)

from asl_workflow_engine.logger import init_logging
from asl_workflow_engine.open_tracing_factory import span_context, inject_span
from asl_workflow_engine.store import (
    create_ASL_store,
    create_executions_store,
    create_history_store,
)
from asl_workflow_engine.task_dispatcher import TaskDispatcher

from asl_workflow_engine.asl_exceptions import *
from asl_workflow_engine.arn import *

try:  # Attempt to use ujson if available https://pypi.org/project/ujson/
    import ujson as json
except:  # Fall back to standard library json
    import json

"""
Stepfunction quota/size limits
https://docs.aws.amazon.com/step-functions/latest/dg/limits.html
https://docs.aws.amazon.com/step-functions/latest/apireference/API_StartExecution.html
https://docs.aws.amazon.com/step-functions/latest/apireference/API_CreateStateMachine.html
"""
MAX_EXECUTION_HISTORY_LENGTH = 25000
MAX_DATA_LENGTH = 262144  # Max length of the input or output JSON string.
MAX_STATE_MACHINE_LENGTH = 1048576  # Max length of the State Machine definition.

def parse_rfc3339_datetime(rfc3339):
    """
    Parse an RFC3339 (https://www.ietf.org/rfc/rfc3339.txt) format string into
    a datetime object which is essentially the inverse operation to
    datetime.now(timezone.utc).astimezone().isoformat()
    We primarily need this in the Wait state so we can compute timeouts etc.
    """
    rfc3339 = rfc3339.strip()  # Remove any leading/trailing whitespace
    if rfc3339[-1] == "Z":
        date = rfc3339[:-1]
        offset = "+00:00"
    else:
        date = rfc3339[:-6]
        offset = rfc3339[-6:]

    if "." not in date:
        date = date + ".0"
    raw_datetime = datetime.strptime(date, "%Y-%m-%dT%H:%M:%S.%f")
    delta = timedelta(hours=int(offset[-5:-3]), minutes=int(offset[-2]))
    if offset[0] == "-":
        delta = -delta
    return raw_datetime.replace(tzinfo=timezone(delta))

def find_state(current_state_machine, current_state):
    """
    Look-up the specified JSON state machine to find the state object with the
    given name. If it can't be found by a simple look-up try to find it in
    a nested (e.g. Parallel or Map) state machine as described below.
    """
    state = current_state_machine.get(current_state)
    if state == None:
        """
        If the state can't be found in the parent state machine search for
        it more deeply using recursive descent, as the specified state might
        actually be in a Parallel branch or Map Iterator state machine.
        Because JSONPath doesn't have a parent operator and we want to get
        the parent States object too we get the full JSONPath string for
        the query then use simple string splits to find the path of that.
        """
        path = get_full_jsonpath(current_state_machine, "$.." + current_state)
        states_path = path.rpartition("['States']")[0]
        if path and states_path:
            branch = apply_jsonpath(current_state_machine, states_path)
            current_state_machine = branch["States"]
            state = current_state_machine.get(current_state)

    return state, current_state_machine

def merge_result(data, context, result, state, output_path=None):
    """
    Boiler plate to apply both ResultPath and OutputPath (or supplied output_path)
    """
    output = apply_resultpath(
        data, result, state.get("ResultPath", "$")
    )

    output_path = output_path if output_path else state.get("OutputPath", "$")
    return apply_path(output, context, output_path)


class StateEngine(object):
    def __init__(self, config):
        """
        """
        self.logger = init_logging(log_name="asl_workflow_engine")
        self.logger.info("Creating StateEngine, using {} JSON parser".format(json.__name__))

        store_url = config["state_engine"]["store_url"]
        
        """
        The ASL store is semantically a dict of dict. The outer dict holds
        ASL objects (also dicts) keyed by the ARN of the ASL State Machine.
        The format of the ASL objects is the same as DescribeStateMachine.
        https://docs.aws.amazon.com/step-functions/latest/apireference/API_DescribeStateMachine.html
        """
        self.asl_store = create_ASL_store(store_url)

        """
        Holds information required by the DescribeExecution API call.
        The executions store is semantically a dict of dict.
        https://docs.aws.amazon.com/step-functions/latest/apireference/API_DescribeExecution.html
        """
        self.executions = create_executions_store(store_url)

        """
        Holds information required by the GetExecutionHistory API call.
        The executions store is semantically a dict of list of immutable dicts.
        https://docs.aws.amazon.com/step-functions/latest/apireference/API_GetExecutionHistory.html
        """
        self.execution_history = create_history_store(store_url)

        """
        The execution_ttl is the time to live, in seconds, for the execution
        metadata used by DescribeExecution (executions) and GetExecutionHistory
        (execution_history). The initial default is one day 60x60x24 = 86400.
        Whether to raise or lower this value will depend on the available memory
        of the Redis server. The execution_ttl is only honoured by the Redis
        store and is ignored by the simple store.
        """
        self.execution_ttl = config["state_engine"]["execution_ttl"]

        """
        Holds the results of Parallel state Branches or Map state Iterations.
        It holds objects keyed by the execution ARN and each object is another
        dict which contains arrays keyed by the Parallel or Map state name.
        """
        self.branch_results = {}

        self.task_dispatcher = TaskDispatcher(self, config)
        # self.event_dispatcher set by EventDispatcher

        """
        Prometheus metrics intended to emulate Stepfunction CloudWatch metrics.
        https://docs.aws.amazon.com/step-functions/latest/dg/procedure-cw-metrics.html
        """
        self.execution_metrics = {}
        metrics_config = config.get("metrics", {})
        if metrics_config.get("implementation", "") == "Prometheus":
            ns = metrics_config.get("namespace", "")
            if ns:
                self.logger.info("Enabling Prometheus Metrics in namespace: " + ns)
                ns += "_"
            else:
                self.logger.info("Enabling Prometheus Metrics")

            self.execution_metrics = {
                "ExecutionTime": Summary(
                    ns + "ExecutionTime",
                    "The interval, in milliseconds, between the time the " +
                    "execution starts and the time it closes."
                ),
                "ExecutionsFailed": Counter(
                    ns + "ExecutionsFailed",
                    "The number of failed executions."
                ),
                "ExecutionsStarted": Counter(
                    ns + "ExecutionsStarted",
                    "The number of started executions."
                ),
                "ExecutionsSucceeded": Counter(
                    ns + "ExecutionsSucceeded",
                    "The number of successfully completed executions."
                ),
                "ExecutionsTimedOut": Counter(
                    ns + "ExecutionsTimedOut",
                    "The number of executions that time out for any reason."
                )
            }

    def log_and_drop(self, message, obj, id):
        """
        Boiler plate to deal with unrecoverable errors that should rarely occur.
        """
        self.logger.error(
            ("StateEngine: " + message + ", dropping the message!").format(obj)
        )
        self.event_dispatcher.acknowledge(id)

    def broadcast_notification(self, execution_arn, execution_detail, context):
        """
        Broadcasts notification of execution status changes to a topic.

        The intention is to act rather like AWS CloudWatch events:
        https://docs.aws.amazon.com/step-functions/latest/dg/cw-events.html
        https://docs.aws.amazon.com/AmazonCloudWatch/latest/events/CloudWatchEventsandEventPatterns.html
        The events are published to the topic/subject:
        asl_workflow_engine/<state_machine_arn>.<status>

        That is to say the asl_workflow_engine topic exchange with a subject
        comprising the state_machine_arn and the status separated by a dot. e.g. 
        asl_workflow_engine/arn:aws:states:local:0123456789:stateMachine:simple_state_machine.SUCCEEDED

        Consumers may subscribe to specific events using the full subject or
        groups of events using wildcards, for example the subscription:
        asl_workflow_engine/arn:aws:states:local:0123456789:stateMachine:simple_state_machine.*
        Subscribes to all notification events published for a given state machine.

        The format of the message body is as described in
        https://docs.aws.amazon.com/step-functions/latest/dg/cw-events.html
        the "detail" field gives access to the "executionArn", "stateMachineArn",
        "input", "output" and other relevant information and provides the same
        information as the DescribeExecution API.
        """

        arn = parse_arn(execution_arn)
        region=arn.get("region", "local")
        account=arn["account"]

        """
        https://docs.aws.amazon.com/step-functions/latest/dg/cw-events.html
        https://docs.aws.amazon.com/AmazonCloudWatch/latest/events/CloudWatchEventsandEventPatterns.html
        https://stackoverflow.com/questions/8556398/generate-rfc-3339-timestamp-in-python
        """
        if not isinstance(execution_detail , dict):  # May be (non JSON) RedisDict
            execution_detail = dict(execution_detail)

        """
        There is an inconsistency in AWS. The DescribeExecution documentation
        has startDate and stopDate described as number and Type: Timestamp, but
        doesn't mention anything about Timestamp units (seconds or millis),
        however boto3 barfs on start_execution, list_execution and
        describe_execution unless timestamp is in epoch seconds e.g. time.time()
        https://docs.aws.amazon.com/step-functions/latest/apireference/API_DescribeExecution.html
        However the EventBridge (CloudWatch Events) example here:
        https://docs.aws.amazon.com/step-functions/latest/dg/cw-events.html#cw-events-execution-succeeded
        illustrates startDate and stopDate as millisecond Timestamps and that
        seems to be the case for real with CW Events too, so we need to
        convert to millis here. Rather than clone the whole dict just for
        those two fields we save their original values and copy them back after
        we broadcast the message.
        """
        saved_startDate = execution_detail["startDate"]
        saved_stopDate = execution_detail["stopDate"]

        if saved_startDate:
            execution_detail["startDate"] = int(saved_startDate * 1000)
        if execution_detail["stopDate"]:
            execution_detail["stopDate"] = int(saved_stopDate * 1000)

        """
        Start an OpenTracing trace for the notification.
        https://opentracing.io/guides/python/tracers/ standard tags are from
        https://opentracing.io/specification/conventions/
        """
        with opentracing.tracer.start_active_span(
            operation_name="StatusChange",
            child_of=span_context("text_map", context.get("Tracer", {}), self.logger),
            tags={
                "component": "state_engine",
                "execution_arn": execution_arn,
            }
        ) as scope:
            carrier = inject_span("text_map", scope.span, self.logger)
            event_time = datetime.now(timezone.utc).astimezone().isoformat()
            cw_event = {
                "version": "0",  # By default, this is set to 0 (zero) in all events.
                "id": str(uuid.uuid4()),  # A unique value is generated for every event.
                "detail-type": "Step Functions Execution Status Change",
                "source": "aws.states",  # Identifies the service that sourced the event.
                "account": account,  # The 12-digit number identifying an AWS account.
                "time": event_time,  # Seems to be in rfc-3339 format in examples.
                "region": region,  # Identifies the AWS region where the event originated.
                "resources": [execution_arn],
                "detail": execution_detail,
            }

            subject = execution_detail["stateMachineArn"] + "." + execution_detail["status"]
            self.event_dispatcher.broadcast(subject, cw_event, carrier_properties=carrier)

        # Copy the original seconds since epoch timestamps back.
        execution_detail["startDate"] = saved_startDate
        execution_detail["stopDate"] = saved_stopDate

    def start_execution(self, state_machine_type, start_state, event):
        """
        Initialise the state machine execution's state, in particular this will
        set any context metadata that hasn't previously been set elsewhere.
        This is mostly only necessary if the execution was started in a
        "low-level" e.g. way by publishing directly to the event queue, whereas
        if the execution was started via the REST API the context metadata should
        already be set and this method's purpose is then simply to update the
        execution history metadata.
        """
        data = event["data"]
        context = event["context"]
        state_machine_arn = context["StateMachine"]["Id"]

        context["State"]["Name"] = start_state

        if "Execution" not in context:
            context["Execution"] = {}
        execution = context["Execution"]

        if "Name" not in execution:
            execution["Name"] = str(uuid.uuid4())

        if "Id" not in execution:
            # Create Id
            # Form executionArn from stateMachineArn and uuid
            arn = parse_arn(state_machine_arn)
            execution_arn = create_arn(
                service="states",
                region=arn.get("region", "local"),
                account=arn["account"],
                resource_type="execution",
                resource=arn["resource"] + ":" + execution["Name"],
            )
            execution["Id"] = execution_arn

        execution_arn = execution["Id"]  # The previous block ensures this is set.

        """
        execution["Input"] holds the initial Step Function input. As this is
        part of the Context Object the "Input" field here is a JSON object.
        https://docs.aws.amazon.com/step-functions/latest/dg/input-output-contextobject.html
        """
        if "Input" not in execution:
            execution["Input"] = data

        if "RoleArn" not in execution:
            """
            The default dummy ARN case shouldn't generally occur unless
            the asl_store "database" has been manually edited because when
            it is updated via the API the roleArn field gets populated.
            """
            asl_item = self.asl_store.get_cached_view(state_machine_arn, {})
            execution["RoleArn"] = asl_item.get(
                "roleArn", "arn:aws:iam:::role/dummy-role/dummy"
            )

        # https://stackoverflow.com/questions/8556398/generate-rfc-3339-timestamp-in-python
        start_time = datetime.now(timezone.utc).astimezone().isoformat()
        if "StartTime" not in execution:
            execution["StartTime"] = start_time

        """
        For the start state set EnteredTime here, which will be reset if
        the message gets redelivered, for other state transitions we will
        set EnteredTime when the transition occurs so if redeliveries
        occur the EnteredTime will reflect the time the state was entered
        not the time the redelivery occurred.
        """
        if "EnteredTime" not in context["State"]:
            context["State"]["EnteredTime"] = start_time
            
        """
        Populate the metadata required by the DescribeExecution API call.
        https://docs.aws.amazon.com/step-functions/latest/apireference/API_DescribeExecution.html
        Note that the "input" and "output" fields of the execution detail and
        history are actually JSON strings not JSON objects.
        It actually took a while to note that subtlety, especially as the AWS
        CLI describe-execution call doesn't actually complain and displayed the
        JSON object that was originally being sent before json.dumps() was added.
        """
        input_as_string = json.dumps(data)
        execution_detail = {
            "executionArn": execution_arn,
            "input": input_as_string,
            "name": execution["Name"],
            "output": None,
            "startDate": time.time(),
            "stateMachineArn": state_machine_arn,
            "status": "RUNNING",
            "stopDate": None,
        }

        """
        Only store execution metadata/history for "STANDARD" not "EXPRESS"
        State Machine workflows.
        """
        if state_machine_type == "STANDARD":
            self.executions[execution_arn] = execution_detail
            self.executions.set_ttl(execution_arn, self.execution_ttl)

            # Don't set the execution_history ttl here, wait until first append.
            self.execution_history[execution_arn] = []

        """
        Still call this for "EXPRESS" State Machine workflows. Initially it will
        return immediately for "EXPRESS" workflows, but we may do extra logging.
        """
        self.update_execution_history(
            state_machine_type,
            execution_arn,
            "ExecutionStarted",
            {"input": input_as_string, "roleArn": execution["RoleArn"]},
        )
        print(self.execution_metrics)
        if self.execution_metrics:
            self.execution_metrics["ExecutionsStarted"].inc(
                {"StateMachineArn": state_machine_arn}
            )

        self.broadcast_notification(execution_arn, execution_detail, context)

    def change_state(self, state_machine_type, state_type, next_state, event):
        """
        Set event's new current state in $$.State.Name to Next state.
        """
        data = event["data"]
        """
        Note that the "output" field of the execution detail and history is a
        JSON string not a JSON object, hence the json.dumps().
        https://docs.aws.amazon.com/step-functions/latest/apireference/API_DescribeExecution.html
        https://docs.aws.amazon.com/step-functions/latest/apireference/API_GetExecutionHistory.html
        """
        output_as_string = json.dumps(data)
        context = event["context"]
        state = context["State"]

        """
        First check next_state is not None. This shouldn't be the case as it is
        invalid ASL, however in the absence of an effective ASL validity checker
        in the ASL Engine at the moment this defensive logic will terminate
        the execution should this situation occur.
        """
        if next_state == None:
            error_message = ("An error occurred while executing the state '{}'. "
                             "Mandatory \"Next\" field is missing."
                             ).format(state["Name"])
            event["data"] = {
                "Error": "States.Runtime",
                "Cause": error_message,
            }
            self.logger.error("States.Runtime: {}".format(error_message))
            self.end_execution(state_machine_type, state_type, event)
            return

        """
        Check if the State's aggregate output has exceeded the 262144
        character quota described in Stepfunction Quotas page.
        https://docs.aws.amazon.com/step-functions/latest/dg/limits.html
        If so then we fail the execution.
        """
        if len(output_as_string) > MAX_DATA_LENGTH:
            error_message = ("An error occurred while executing the state '{}'. "
                             "A result with a size exceeding the maximum "
                             "number of characters service limit "
                             "was returned.").format(state["Name"])
            event["data"] = {
                "Error": "States.DataLimitExceeded",
                "Cause": error_message,
            }
            self.logger.error("States.DataLimitExceeded: {}".format(error_message))
            self.end_execution(state_machine_type, state_type, event)
            return

        self.update_execution_history(
            state_machine_type,
            context["Execution"]["Id"],
            state_type + "StateExited",
            {"output": output_as_string, "name": state["Name"]},
        )
        # Update the state with the next state's name and erase any retry info.
        state["Name"] = next_state
        if "RetryCount" in state:
            del state["RetryCount"]
        if "RetryTimeout" in state:
            del state["RetryTimeout"]

        # https://stackoverflow.com/questions/8556398/generate-rfc-3339-timestamp-in-python
        state["EnteredTime"] = datetime.now(timezone.utc).astimezone().isoformat()
        self.event_dispatcher.publish(event)

    def end_execution(self, state_machine_type, state_type, event):
        """
        End the state machine execution and update execution metadata.
        """
        data = event["data"]
        """
        Note that the "output" field of the execution detail and history is a
        JSON string not a JSON object, hence the json.dumps().
        https://docs.aws.amazon.com/step-functions/latest/apireference/API_DescribeExecution.html
        https://docs.aws.amazon.com/step-functions/latest/apireference/API_GetExecutionHistory.html
        """
        output_as_string = json.dumps(data)
        context = event["context"]
        state = context["State"]
        execution_arn = context["Execution"]["Id"]

        execution_failed = isinstance(data, dict) and data.get("Error")

        # Stepfunctions don't transition to StateExited if the execution fails.
        if not execution_failed:
            self.update_execution_history(
                state_machine_type,
                execution_arn,
                state_type + "StateExited",
                {"output": output_as_string, "name": state["Name"]},
            )

        if state_machine_type == "STANDARD":
            execution_detail = self.executions[execution_arn]
            state_machine_arn = execution_detail["stateMachineArn"]
        else:
            """
            For "EXPRESS" workflows we don't store the execution metadata in
            self.executions, however the missing information for the current
            execution is actually available in the execution context or can
            be derived from the execution ARN.
            """
            # Derive missing fields from execution_arn
            split = execution_arn.rpartition(':')
            arn = parse_arn(split[0])
            arn["resource_type"] = "stateMachine"
            state_machine_arn = create_arn(arn)
            name = split[2]

            execution = context["Execution"]
            start_date = parse_rfc3339_datetime(execution["StartTime"]).timestamp()

            execution_detail = {
                "executionArn": execution_arn,
                "input": json.dumps(execution["Input"]),
                "name": name,
                "output": None,
                "startDate": start_date,
                "stateMachineArn": state_machine_arn,
                "status": "RUNNING",
                "stopDate": None,
            }

        execution_detail["stopDate"] = time.time()

        with opentracing.tracer.start_active_span(
            operation_name="StartExecution:ExecutionEnding",
            child_of=span_context("text_map", context.get("Tracer", {}), self.logger),
            tags={
                "component": "state_engine",
                "execution_arn": execution_arn
            }
        ) as scope:
            if execution_failed:
                opentracing.tracer.active_span.set_tag("error", True)
                opentracing.tracer.active_span.log_kv(
                    {
                        "message": "ExecutionFailed",
                    }
                )
                opentracing.tracer.active_span.set_tag("status", "FAILED")
                execution_detail["status"] = "FAILED"
                """
                As per the DescribeExecution API, "output" is set only if the
                execution succeeds. If the execution fails, this field is null. 
                https://docs.aws.amazon.com/step-functions/latest/apireference/API_DescribeExecution.html
                """
                execution_detail["output"] = None
                self.update_execution_history(
                    state_machine_type,
                    execution_arn,
                    "ExecutionFailed",
                    {"cause": data.get("Cause"), "error": data.get("Error")}
                )
                self.logger=init_logging(log_name="asl_execution_history")
                self.logger.info(self.execution_history)
                if self.execution_metrics:
                    if data.get("Error") == "States.Timeout":
                        self.execution_metrics["ExecutionsTimedOut"].inc(
                            {"StateMachineArn": state_machine_arn}
                        )

                    self.execution_metrics["ExecutionsFailed"].inc(
                        {"StateMachineArn": state_machine_arn}
                    )
            else:
                opentracing.tracer.active_span.set_tag("status", "SUCCEEDED")
                execution_detail["status"] = "SUCCEEDED"
                execution_detail["output"] = output_as_string
                self.update_execution_history(
                    state_machine_type,
                    execution_arn,
                    "ExecutionSucceeded",
                    {"output": output_as_string}
                )

                if self.execution_metrics:
                    self.execution_metrics["ExecutionsSucceeded"].inc(
                        {"StateMachineArn": state_machine_arn}
                    )
        
        # Tidy up self.branch_results for current execution_arn.
        if execution_arn in self.branch_results:
            del self.branch_results[execution_arn]

        if self.execution_metrics:
            duration = (execution_detail["stopDate"] - 
                        execution_detail["startDate"]) * 1000.0
            self.execution_metrics["ExecutionTime"].observe(
                {"StateMachineArn": state_machine_arn}, duration 
            )

        self.broadcast_notification(execution_arn, execution_detail, context)

    def update_execution_history(
            self, state_machine_type, execution_arn, update_type, details
        ):

        """
        TODO For "EXPRESS" State Machines we might actually want to log with a
        higher level than debug. "STANDARD" State Machines store execution
        history and other execution metadata but "EXPRESS" State Machines don't.
        However Express Workflows send execution history to CloudWatch Logs.
        https://aws.amazon.com/blogs/compute/new-express-workflows-for-aws-step-functions/
        """
        if state_machine_type == "EXPRESS":
            self.logger.debug("{} {} {}".format(execution_arn, update_type, details))
            return

        """
        Store the execution history information in a way that will be accessible
        to the REST API.

        Only do DEBUG logging, as the details object is likely to be expensive
        and in any case the idea is to get execution history via the REST API
        https://docs.aws.amazon.com/step-functions/latest/apireference/API_GetExecutionHistory.html
        """
        self.logger.debug("{} {} {}".format(execution_arn, update_type, details))

        """
        self.executions.get(execution_arn) == None should only really happen if
        the StateEngine has failed and been restarted and we are handling a
        redelivered message. When we add code IDC to persist execution metadata
        state we should hopefully be able to avoid the following condition upon
        StateEngine restart.
        """
        if self.executions.get(execution_arn) == None:
            self.logger.warning(
                "StateEngine: update_execution_history: Execution {} does not "
                "exist, probably due to StateEngine restart. Some history "
                "metadata has been lost!".format(execution_arn))

            # Derive missing fields from execution_arn
            split = execution_arn.rpartition(':')
            arn = parse_arn(split[0])
            arn["resource_type"] = "stateMachine"
            state_machine_arn = create_arn(arn)
            name = split[2]

            self.executions[execution_arn] = {
                "executionArn": execution_arn,
                "input": None,
                "name": name,
                "output": None,
                "startDate": time.time(),
                "stateMachineArn": state_machine_arn,
                "status": "RUNNING",
                "stopDate": None,
            }
            self.execution_history[execution_arn] = []

        history = self.execution_history[execution_arn]
        """
        Events are numbered sequentially, starting at one.
        https://docs.aws.amazon.com/step-functions/latest/apireference/API_HistoryEvent.html
        """
        id = len(history) + 1
        if "StateEntered" in update_type:
            details_type = "stateEnteredEventDetails"
        elif "StateExited" in update_type:
            details_type = "stateExitedEventDetails"
        else:
            details_type = update_type[0].lower() + update_type[1:] + "EventDetails"

        history.append(
            {
                "timestamp": time.time(),
                details_type: details,
                "type": update_type,
                "id": id,
                "previousEventId": id - 1,
            }
        )

        if id == 1:
            """
            Defer setting execution_history ttl to here, as we now have an item
            in the execution_history store to associate the ttl with.
            """
            self.execution_history.set_ttl(execution_arn, self.execution_ttl)

    def notify(self, event, id):
        """
        ------------------------------------------------------------------------
        This method is the main entry point to the StateEngine. It is called by
        the EventDispatcher when State events are available on the event queue.
        ------------------------------------------------------------------------

        :item event: Describes the data, current state and the ASL State Machine
        :type event: dict as described below. N.B. note that the expected type
         is a dictionary i.e. a JSON object and NOT a JSON string!
        :item id: The ID of the event as given by the EventDispatcher, it is
         primarily used for acknowledging the event.
        :type id: A string representing the event ID, it may just be a number.

        The event dict passed to the notify method has the following format:

        {
	        "data": <Object representing application data>,
	        "context": <Object representing application context>,
        }

        The application context is described in the AWS documentation:
        https://docs.aws.amazon.com/step-functions/latest/dg/input-output-contextobject.html 

        {
	        "Execution": {
		        "Id": <String>,
		        "Input": <Object>,
		        "StartTime": <String Format: ISO 8601>
	        },
	        "State": {
		        "EnteredTime": <String Format: ISO 8601>,
		        "Name": <String>,
		        "RetryCount": <Number>
	        },
	        "StateMachine": {
		        "Id": <String>,
		        "Definition": <Object representing ASL state machine>
	        },
	        "Task": {
		        "Token": <String>
	        }
        }

        The most important paths for state traversal are:
        $$.State.Name = the current state
        $$.StateMachine.Definition = (optional) contains the complete ASL state machine
        $$.StateMachine.Id = a unique reference to an ASL state machine
        """
        context = event.get("context")
        if context == None:
            self.log_and_drop("event {} has no $.context field", event, id)
            return

        state_machine = context.get("StateMachine")
        if state_machine == None:
            self.log_and_drop("event {} has no $.context.StateMachine field", event, id)
            return

        state_machine_arn = state_machine.get("Id")
        if not state_machine_arn:
            self.log_and_drop("event {} has no $.context.StateMachine.Id field", event, id)
            return

        """
        If ASL is present in optional state_machine["Definition"] store that as
        if it were a CreateStateMachine API call. TODO the roleArn added here is
        not a valid IAM role ARN. This way of adding State Nachines "by value"
        embedded in the context was mainly added to enable development of some
        features prior to the addition of the REST API, so it might eventually
        be deprecated, alternatively the approach could be enhanced to allow
        a valid roleArn to be added from config or indeed be embedded in the
        context in a similar way to how state_machine["Definition"] is.
        """
        if "Definition" in state_machine:
            arn = parse_arn(state_machine_arn)
            creation_date = time.time()
            self.asl_store[state_machine_arn] = {
                "creationDate": creation_date,
                "definition": state_machine["Definition"],
                "name": arn["resource"],
                "roleArn": "arn:aws:iam:::role/dummy-role/dummy",
                "stateMachineArn": state_machine_arn,
                "updateDate": creation_date,
                "status": "ACTIVE",
                "type": "STANDARD",
            }
            del state_machine["Definition"]


        asl_item = self.asl_store.get_cached_view(state_machine_arn, {})
        ASL = asl_item.get("definition")
        if not ASL:
            """
            Dropping the message is possibly not the right thing to do in this
            scenario as we *could* simply add the State Machine and retry. OTOH
            if a State Machine is deleted executions should also be deleted.
            """
            self.log_and_drop("State Machine {} does not exist", state_machine_arn, id)
            return

        """
        Record if the State Machine for this execution is "STANDARD" or "EXPRESS".
        If "EXPRESS" certain features, like recording execution descriptions and
        history, will be disabled as per the Stepfunction API.
        """
        state_machine_type = asl_item.get("type")

        """
        https://states-language.net/spec.html#data
        https://docs.aws.amazon.com/step-functions/latest/dg/concepts-state-machine-data.html

        The interpreter passes data between states to perform calculations or to
        dynamically control the state machine’s flow. All such data MUST be
        expressed in JSON.

        When a state machine is started, the caller can provide an initial JSON
        text as input, which is passed to the machine's start state as input. If
        no input is provided, the default is an empty JSON object, {}. As each
        state is executed, it receives a JSON text as input and can produce
        arbitrary output, which MUST be a JSON text. When two states are linked
        by a transition, the output from the first state is passed as input to
        the second state. The output from the machine's terminal state is
        treated as its output. 
        """
        data = event.get("data", {})  # Note default to empty JSON object


        # Determine the current state from $$.State.Name.
        # Use get() defaults to cater for State or Name being initially unset.
        if not context.get("State"):
            context["State"] = {"Name": None}
        current_state = context["State"].get("Name")

        if not current_state:
            """
            If current_state is uninitialised it means we are the Start State.
            If so initialise unset context fields and start OpenTracing span.
            """
            current_state = ASL["StartAt"]
            self.start_execution(state_machine_type, current_state, event)

            with opentracing.tracer.start_active_span(
                operation_name="StartExecution:ExecutionStarting",
                child_of=span_context("text_map", context.get("Tracer", {}), self.logger),
                tags={
                    "component": "state_engine",
                    "execution_arn": context["Execution"]["Id"]
                }
            ) as scope:
                context["Tracer"] = inject_span("text_map", scope.span, self.logger)

        """
        States in any given “States” field can transition only to each other,
        and no state outside of that “States” field can transition into it.
        That is to say a parent state machine cannot transition directly to
        a state within a Parallel branch or Map Iterator state machine nor
        can states within those state machines directly transition to states
        outside of their own “States” field. TODO use current_state_machine to
        check that state transitions only occur within the correct "States".
        """
        state, current_state_machine = find_state(ASL["States"], current_state)
        if state == None:  # state should be valid by this point
            self.log_and_drop("State {} does not exist", current_state, id)
            return

        # Determine the ASL state type of the current state.
        state_type = state["Type"]

        # print("state_machine_arn = " + state_machine_arn)
        # print("current_state = " + current_state)
        # print("state = " + str(state))
        # print("data = " + str(data))
        # print("event id = " + str(id))
 

        # ----------------------------------------------------------------------
        def handle_error(error_type, error_message):
            """
            https://states-language.net/spec.html#retrying-after-error Task
            States, Parallel States, and Map States MAY have a field named
            “Retry”, whose value MUST be an array of objects, called Retriers.
            """
            retry_matched = False
            catch_matched = False

            if state_machine_type == "STANDARD":
                execution_arn = context["Execution"]["Id"]
                id = len(self.execution_history[execution_arn])
                boiler_plate = (
                    "An error occurred while executing the state "
                    "'{}' (entered at the event id #{}). "
                ).format(current_state, id)
            else:
                boiler_plate = (
                    "An error occurred while executing the state '{}'. "
                ).format(current_state)

            """
            An execution failed due to some exception that could not be
            processed. Often these are caused by errors at runtime, such
            as attempting to apply InputPath or OutputPath on a null
            JSON payload. A States.Runtime error is not retriable, and
            will always cause the execution to fail.
            A retry or catch on States.ALL will not catch States.Runtime errors. 
            """
            runtime_error = (error_type == "States.Runtime")

            retry = state.get("Retry")
            if not runtime_error and retry and isinstance(retry, list):
                for retrier in retry:
                    """
                    Each Retrier MUST contain a field named “ErrorEquals”
                    whose value MUST be a non-empty array of Strings,
                    which match Error Names.

                    When a state reports an error, the interpreter scans
                    through the Retriers and, when the Error Name appears
                    in the value of a Retrier’s “ErrorEquals” field,
                    implements the retry policy described in that Retrier.

                    How to correctly handle States.TaskFailed is a little
                    unclear, but the AWS Error Handling document here:
                    https://docs.aws.amazon.com/step-functions/latest/dg/concepts-error-handling.html
                    says "States.TaskFailed, which matches any error that
                    a Lambda function outputs." which suggests that if
                    error_type has *any* value and States.TaskFailed is
                    present in the ErrorEquals array then we should match.
                    """
                    error_equals = retrier.get("ErrorEquals")
                    if (
                        error_type in error_equals
                        or "States.TaskFailed" in error_equals
                        or (len(error_equals) == 1
                            and error_equals[0] == "States.ALL")
                    ):
                        # Default values taken from ASL specification.
                        interval_seconds = retrier.get("IntervalSeconds", 1)
                        max_attempts = retrier.get("MaxAttempts", 3)
                        backoff_rate = retrier.get("BackoffRate", 2.0)
                        if backoff_rate < 1.0:
                            backoff_rate = 1.0

                        retries = context["State"].get("RetryCount", 0)
                        if retries < max_attempts:
                            retry_matched = True
                            timeout = interval_seconds * (backoff_rate ** retries)
                            retries += 1
                            context["State"]["RetryCount"] = retries
                            context["State"]["RetryTimeout"] = timeout * 1000
                            context["State"]["EnteredTime"] = (
                                datetime.fromtimestamp(
                                    time.time() + timeout, timezone.utc
                                ).astimezone().isoformat()
                            )
                            """
                            Republish the Task state event with the new
                            RetryCount and RetryTimeout set. We also adjust
                            EnteredTime above. The ASL spec is unclear on
                            this, but if the error is due to States.Timeout
                            and we retry it would seem logical to reset,
                            otherwise when we retry the Task will timeout
                            immediately, similarly we add the RetryTimeout
                            to the EnteredTime as we don't actually start
                            retrying any Task resource until after
                            RetryTimeout has expired.
                            TODO - check the behaviour of AWS Step Functions
                            to see what they actually do in this scenario.
                            """
                            self.event_dispatcher.publish(event)

                        break

            """
            https://states-language.net/spec.html#fallback-states
            When a state reports an error and either there is no Retrier, or
            retries have failed to resolve the error, the interpreter scans
            through the Catchers in array order, and when the Error Name appears
            in the value of a Catcher’s “ErrorEquals” field, transitions the
            machine to the state named in the value of the “Next” field.

            When a state has both “Retry” and “Catch” fields, the interpreter
            uses any appropriate Retriers first and only applies the a matching
            Catcher transition if the retry policy fails to resolve the error.
            """
            catch = state.get("Catch")
            if not runtime_error and not retry_matched and catch and isinstance(catch, list):
                for catcher in catch:
                    """
                    Each Catcher MUST contain a field named “ErrorEquals”,
                    specified exactly as with the Retrier “ErrorEquals” field,
                    and a field named “Next” whose value MUST be a string
                    exactly matching a State Name.

                    When a state reports an error and either there is no
                    Retrier, or retries have failed to resolve the error,
                    the interpreter scans through the Catchers in array order,
                    and when the Error Name appears in the value of a Catcher’s 
                    “ErrorEquals” field, transitions the machine to the state
                    named in the value of the “Next” field.
                    """
                    error_equals = catcher.get("ErrorEquals")
                    if (
                        error_type in error_equals
                        or "States.TaskFailed" in error_equals
                        or (len(error_equals) == 1
                            and error_equals[0] == "States.ALL")
                    ):
                        catch_matched = True
                        """
                        When a state reports an error and it matches a
                        Catcher, causing a transfer to another state, the
                        state’s Result (and thus the input to the state
                        identified in the Catcher’s “Next” field) is a JSON
                        object, called the Error Output. The Error Output
                        MUST have a string-valued field named “Error”,
                        containing the Error Name. It SHOULD have a string
                        valued field named “Cause”, containing human-readable
                        text about the error.
                        """
                        result = {"Error": error_type}
                        if error_message:
                            result["Cause"] = boiler_plate + error_message

                        """
                        A Catcher MAY have an “ResultPath” field, which works
                        exactly like a state’s top-level “ResultPath”, and
                        may be used to inject the Error Output into the
                        state’s original raw input to create the input for the
                        Catcher’s “Next” state. The default value, if the
                        “ResultPath” field is not provided, is “$”, meaning
                        that the output consists entirely of the Error Output. 
                        """
                        event["data"] = merge_result(data, context, result, catcher, "$")

                        self.change_state(
                            state_machine_type,
                            state_type,
                            catcher.get("Next"),
                            event
                        )
                        break

            """
            When a state reports an error, the default course of action for the
            interpreter is to fail the whole state machine, so if no retriers or
            catchers match then fail the execution.
            """
            if not retry_matched and not catch_matched:
                result = {"Error": error_type}
                if error_message:
                    result["Cause"] = boiler_plate + error_message

                event["data"] = result
                self.end_execution(state_machine_type, state_type, event)

        # ----------------------------------------------------------------------


        def handle_terminal_state(state_type, event, id=None):
            """
            This function handles the boilerplate needed for terminal states.
            """
            if "Branch" in context["State"]:
                execution_arn = context["Execution"]["Id"]
                self.update_execution_history(
                    state_machine_type,
                    execution_arn,
                    state_type + "StateExited",
                    {"output": json.dumps(event["data"]), "name": current_state},
                )
                asl_state_collect_results(state_type)
            else:
                self.end_execution(state_machine_type, state_type, event)
                if id != None:
                    self.event_dispatcher.acknowledge(id)

        """
        Define nested functions as handlers for each supported ASL state type.
        Using nested functions so we can use the context extracted in notify.

        That the methods are prefixed with "asl_state_" is a mitigation against
        accidentally or deliberately placing an invalid State type in the ASL.
        """
        def asl_state_Pass():
            """
            https://states-language.net/spec.html#pass-state
            https://docs.aws.amazon.com/step-functions/latest/dg/amazon-states-language-pass-state.html

            The Pass State (identified by "Type":"Pass") simply passes its input
            to its output, performing no work. Pass States are useful when
            constructing and debugging state machines.
            """
            try:
                input = apply_path(data, context, state.get("InputPath", "$"))

                """
                https://states-language.net/spec.html#using-paths

                If the “Parameters” field is provided, its value, after the
                extraction and embedding, becomes the effective input.
                """
                parameters = evaluate_payload_template(
                    input, context, state.get("Parameters")
                )

                """
                A Pass State MAY have a field named “Result”. If present, its
                value is treated as the output of a virtual task, and placed as
                prescribed by the “ResultPath” field, if any, to be passed on
                to the next state.

                If “Result” is not provided, the output is the input. Thus if
                neither “Result” nor “ResultPath” are provided, the Pass state
                copies its input through to its output.
                """
                result = state.get(
                    "Result", parameters
                ) # Default is the "effective input" as per ASL spec.

                # Pass state applies ResultPath to "raw input"
                event["data"] = merge_result(data, context, result, state)

                if state.get("End"):
                    handle_terminal_state(state_type, event, id)
                else:
                    self.change_state(
                        state_machine_type,
                        state_type,
                        state.get("Next"),
                        event
                    )
                    self.event_dispatcher.acknowledge(id)
            except IntrinsicFailure as e:
                handle_error("States.IntrinsicFailure", str(e))
                self.event_dispatcher.acknowledge(id)
            except PathMatchFailure as e:
                handle_error("States.Runtime", str(e))
                self.event_dispatcher.acknowledge(id)
            except ResultPathMatchFailure as e:
                handle_error("States.ResultPathMatchFailure", str(e))
                self.event_dispatcher.acknowledge(id)

        def asl_state_Task_delegate():
            """
            This function represents the real Task state, it is delegated to by
            the asl_state_Task function when any retry delay interval has expired.

            https://states-language.net/spec.html#task-state
            https://docs.aws.amazon.com/step-functions/latest/dg/amazon-states-language-task-state.html
            """

            """
            It's important for the on_response function to be nested as we want
            the event, state and id to be wrapped in its closure, to be used when
            the service integrated to the Task *actually* returns its result.
            """
            def on_response(result):
                """
                The use of the "errorType" field to report an error invoking a
                Task isn't actually mentioned in the ASL specification, but this
                is the pattern adopted for error messages returned by AWS Lambda
                so is almost certainly a fair starting point.
                https://docs.aws.amazon.com/lambda/latest/dg/python-exceptions.html
                https://docs.aws.amazon.com/lambda/latest/dg/nodejs-prog-mode-exceptions.html
                https://docs.aws.amazon.com/lambda/latest/dg/java-exceptions.html
                """
                error_type = result.get("errorType")
                if error_type:
                    if error_type == "States.DataLimitExceeded":
                        error_message = ("The state/task '{}' returned a result "
                                         "with a size exceeding the maximum "
                                         "number of characters service limit.").format(
                                            state.get("Resource")
                                         )
                    else:
                        error_message = result.get("errorMessage", "")
                    self.logger.error("{}: {}".format(error_type, error_message))
                    handle_error(error_type, error_message)
                    self.event_dispatcher.acknowledge(id)
                else:  # No error
                    try:
                        """
                        https://states-language.net/spec.html#using-paths

                        The value of "ResultSelector" MUST be a Payload Template,
                        whose input is the result, and whose payload replaces
                        and becomes the effective result.
                        """
                        result = evaluate_payload_template(
                            result, context, state.get("ResultSelector")
                        )

                        # Task state applies ResultPath to "raw input"
                        event["data"] = merge_result(data, context, result, state)

                        if state.get("End"):
                            handle_terminal_state(state_type, event, id)
                        else:
                            self.change_state(
                                state_machine_type,
                                state_type,
                                state.get("Next"),
                                event
                            )
                            self.event_dispatcher.acknowledge(id)
                    except IntrinsicFailure as e:
                        handle_error("States.IntrinsicFailure", str(e))
                        self.event_dispatcher.acknowledge(id)
                    except PathMatchFailure as e:
                        handle_error("States.Runtime", str(e))
                        self.event_dispatcher.acknowledge(id)
                    except ResultPathMatchFailure as e:
                        handle_error("States.ResultPathMatchFailure", str(e))
                        self.event_dispatcher.acknowledge(id)


            try:
                input = apply_path(data, context, state.get("InputPath", "$"))

                """
                The Task State (identified by "Type":"Task") causes the interpreter
                to execute the work identified by the state’s “Resource” field.

                A Task State MUST include a “Resource” field, whose value MUST
                be a URI that uniquely identifies the specific task to execute.
                The States language does not constrain the URI scheme nor any
                other part of the URI.
                """
                resource_arn = state.get("Resource")

                """
                https://states-language.net/spec.html#using-paths

                If the “Parameters” field is provided, its value, after the
                extraction and embedding, becomes the effective input.
                """
                parameters = evaluate_payload_template(
                    input, context, state.get("Parameters")
                )

                """
                Tasks can optionally specify timeouts. Timeouts (the
                “TimeoutSeconds” and “HeartbeatSeconds” fields) are specified
                in seconds and MUST be positive integers. If provided, the
                “HeartbeatSeconds” interval MUST be smaller than the
                “TimeoutSeconds” value.

                If the state runs longer than the specified timeout, or if
                more time than the specified heartbeat elapses between
                heartbeats from the task, then the interpreter fails the state
                with a States.Timeout Error Name.

                Get the execution and state entry timestamps and the execution
                and state timeout values (if present) and use those to compute
                the Task expiry time that will be checked against the current
                time each heartbeat and will eventually result in the Task
                returning a States.Timeout Error if it expires before it
                completes. The default State Machine execution timeout is
                1 year 60*60*24*365 = 31536000
                https://docs.aws.amazon.com/step-functions/latest/dg/limits.html
                The default Task state timeout is more confusing as this document
                https://docs.aws.amazon.com/step-functions/latest/dg/amazon-states-language-task-state.html says: "If not provided, the default value is 99999999" whereas
                https://states-language.net/spec.html#task-state says "If not
                provided, the default value of “TimeoutSeconds” is 60."
                """
                current_timestamp = time.time()

                start_time = context["Execution"].get("StartTime")
                execution_timestamp = parse_rfc3339_datetime(start_time).timestamp()
                execution_timeout = ASL.get("TimeoutSeconds", 31536000)
                t1 = (execution_timestamp + execution_timeout - current_timestamp) * 1000
                # Negative timeouts could occur if messages are redelivered or
                # backlogged so we set them to 0 if that happens.
                t1 = t1 if t1 > 0 else 0

                entered_time = context["State"].get("EnteredTime")
                state_timestamp = parse_rfc3339_datetime(entered_time).timestamp()

                state_timeout = state.get("TimeoutSecondsPath")
                if state_timeout:
                    state_timeout = apply_path(data, context, state_timeout)
                    state_timeout = state_timeout if isinstance(state_timeout, int) else 0
                else:
                    state_timeout = state.get("TimeoutSeconds", 99999999)
                    #state_timeout = state.get("TimeoutSeconds", 60)  # 60s default
                    #state_timeout = state.get("TimeoutSeconds", 10)  # 10s default

                t2 = (state_timestamp + state_timeout - current_timestamp) * 1000
                # Negative timeouts could occur if messages are redelivered or
                # backlogged so we set them to 0 if that happens.
                t2 = t2 if t2 > 0 else 0

                # The computed timeout is the lowest of the Execution timeout
                # (t1) or the State timeout (t2). 
                timeout = t1 if t1 < t2 else t2

                self.task_dispatcher.execute_task(
                    resource_arn, parameters, on_response, timeout, context
                )
            except IntrinsicFailure as e:
                handle_error("States.IntrinsicFailure", str(e))
                self.event_dispatcher.acknowledge(id)
            except PathMatchFailure as e:
                handle_error("States.Runtime", str(e))
                self.event_dispatcher.acknowledge(id)

        def asl_state_Task():
            """
            This function delegates to the real Task state implementation
            asl_state_Task_delegate when any retry timeout has expired.
            """
            retry_timeout = context["State"].get("RetryTimeout", 0)
            self.event_dispatcher.set_timeout(asl_state_Task_delegate, retry_timeout)

        def asl_state_Choice():
            """
            https://states-language.net/spec.html#choice-state
            https://docs.aws.amazon.com/step-functions/latest/dg/amazon-states-language-choice-state.html

            A Choice state (identified by "Type":"Choice") adds branching logic
            to a state machine.

            TODO - Question, adding the option to perform a regex test would
            seem to be a useful thing to be able to do. This is not defined in
            the ASL specification but is useful so what should we do?
            

            InputPath & OutputPath are allowed (but unusual) in Choice states.
            https://states-language.net/spec.html#statetypes
            """
            try:
                input = apply_path(data, context, state.get("InputPath", "$"))
            except PathMatchFailure as e:
                handle_error("States.Runtime", str(e))
                self.event_dispatcher.acknowledge(id)
                return

            """
            The choose function implements the actual choice logic. We must
            first extract the Variable field and use its value as JSONPath to
            scan the effective input for the actual value the we wish to match.
            """
            def choose(choice):
                """
                "Variable" field may be None for And, Or, Not choice rules
                The throw_exception_on_failed_match flag provides a way to
                test for a Variable/path actually being present.
                """
                path_match_failed = False
                try:
                    variable = apply_path(
                        input,
                        context,
                        choice.get("Variable"),
                        throw_exception_on_failed_match=True
                    )
                except PathMatchFailure:
                    variable = False
                    path_match_failed = True

                next = choice.get("Next", True)


                def isnumber(x):
                    # General test for numeric - any number x zero is zero
                    try:
                        return not isinstance(x, bool) and 0 == x * 0
                    except:
                        return False

                def next_if(variable, op, value, comp_type):
                    # Boiler plate test
                    if (
                        isinstance(variable, comp_type)
                        and isinstance(value, comp_type)
                        and op(variable, value)
                    ):
                        return next

                def next_if_numeric(variable, op, value):
                    # Boiler plate test
                    if (
                        isnumber(variable)
                        and isnumber(value)
                        and op(variable, value)
                    ):
                        return next

                def next_if_timestamp(variable, op, value):
                    """
                    Boiler plate test. The ASL spec. is a little vague on
                    timestamps. The approach we take here is to parse the
                    rfc3339 string into an epoch timestamp and perform the
                    comparisons on those. The thinking with that is because
                    different rfc3339 representations can resolve to the
                    same actual time e.g. a representation in Zulu or local
                    time plus offset can both refer to the same time.
                    """
                    variable = parse_rfc3339_datetime(variable).timestamp()
                    value = parse_rfc3339_datetime(value).timestamp()
                    if op(variable, value):
                        return next
    
                def asl_choice_And(value):
                    if all(choose(ch) for ch in value):
                        return next

                def asl_choice_Or(value):
                    if any(choose(ch) for ch in value):
                        return next

                def asl_choice_Not(value):
                    if (not choose(value)):
                        return next

                def asl_choice_BooleanEquals(value):
                    return next_if(variable, operator.eq, value, bool)

                def asl_choice_NumericEquals(value):
                    return next_if_numeric(variable, operator.eq, value)

                def asl_choice_NumericGreaterThan(value):
                    return next_if_numeric(variable, operator.gt, value)

                def asl_choice_NumericGreaterThanEquals(value):
                    return next_if_numeric(variable, operator.ge, value)

                def asl_choice_NumericLessThan(value):
                    return next_if_numeric(variable, operator.lt, value)

                def asl_choice_NumericLessThanEquals(value):
                    return next_if_numeric(variable, operator.le, value)

                def asl_choice_StringEquals(value):
                    return next_if(variable, operator.eq, value, str)

                def asl_choice_CaseInsensitiveStringEquals(value):
                    # Not covered in ASL spec. but useful and trivial to handle.
                    return next_if(variable.lower(), operator.eq, value.lower(), str)

                def asl_choice_StringGreaterThan(value):
                    return next_if(variable, operator.gt, value, str)

                def asl_choice_StringGreaterThanEquals(value):
                    return next_if(variable, operator.ge, value, str)

                def asl_choice_StringLessThan(value):
                    return next_if(variable, operator.lt, value, str)

                def asl_choice_StringLessThanEquals(value):
                    return next_if(variable, operator.le, value, str)

                def asl_choice_StringMatches(value):
                    # Change the \ escape to fnmatch [seq] escape and also
                    # escape [ to allow things like a literal [hello]
                    value = value.replace("[", "[[]").replace("\*", "[*]")
                    if fnmatch.fnmatch(variable, value):
                        return next

                def asl_choice_TimestampEquals(value):
                    return next_if_timestamp(variable, operator.eq, value)

                def asl_choice_TimestampGreaterThan(value):
                    return next_if_timestamp(variable, operator.gt, value)

                def asl_choice_TimestampGreaterThanEquals(value):
                    return next_if_timestamp(variable, operator.ge, value)

                def asl_choice_TimestampLessThan(value):
                    return next_if_timestamp(variable, operator.lt, value)

                def asl_choice_TimestampLessThanEquals(value):
                    return next_if_timestamp(variable, operator.le, value)

                def asl_choice_IsBoolean(value):
                    if not path_match_failed and (isinstance(variable, bool) == value):
                        return next

                def asl_choice_IsNull(value):
                    if ((variable is None) == value):
                        return next

                def asl_choice_IsNumeric(value):
                    if (isnumber(variable) == value):
                        return next

                def asl_choice_IsString(value):
                    if (isinstance(variable, str) == value):
                        return next

                def asl_choice_IsPresent(value):
                    if (not path_match_failed == value):
                        return next

                def asl_choice_IsTimestamp(value):
                    try:
                        parse_rfc3339_datetime(variable)
                        if (value):
                            return next
                    except Exception:
                        if (not value):
                            return next

                for key in choice:
                    """
                    Determine the ASL choice operator of the current choice and
                    use that to dynamically invoke the appropriate choice handler.
                    """
                    try:
                        value = choice[key]
                        key = "asl_choice_" + key
                        if key.endswith("Path"):  # Handle variable to variable comparison
                            # Slice off "Path" suffix and get value from path
                            key = key[:-4]
                            value = apply_path(data, context, value)

                        next_state = locals().get(key, lambda: None)(value)
                    except Exception:
                        next_state = None

                    if next_state:
                        return next_state

            """
            A Choice state state MUST have a “Choices” field whose value is a
            non-empty array. Each element of the array is called a Choice Rule -
            an object containing a comparison operation and a “Next” field,
            whose value MUST match a state name.
            """
            choices = state.get("Choices", [])  # Sets to [] if key not present
            
            """
            The interpreter attempts pattern-matches against the Choice Rules in
            array order and transitions to the state specified in the “Next”
            field on the first Choice Rule where there is an exact match between
            the input value and a member of the comparison-operator array.
            """
            for choice in choices:
                next_state = choose(choice)
                if next_state:
                    break

            """
            Choice states MAY have a “Default” field, which will execute if none
            of the Choice Rules match. Using state.get("Default") will set the
            value to None if the "Default" field is not present.
            """
            next_state = next_state if next_state else state.get("Default")

            try:
                event["data"] = apply_path(input, context, state.get("OutputPath", "$"))
            except PathMatchFailure as e:
                handle_error("States.Runtime", str(e))
                self.event_dispatcher.acknowledge(id)
                return

            """
            The interpreter will raise a run-time States.NoChoiceMatched error
            if a “Choice” state fails to match a Choice Rule and no “Default”
            transition was specified. 
            """
            if next_state:
                self.change_state(
                    state_machine_type,
                    state_type,
                    next_state,
                    event
                )
            else:
                message = "Choice state {} failed to find a match for the condition field extracted from its input".format(current_state)
                self.logger.error("States.NoChoiceMatched {}".format(message))
                handle_error("States.NoChoiceMatched", message)

            self.event_dispatcher.acknowledge(id)

        def asl_state_Wait():
            """
            https://states-language.net/spec.html#wait-state
            https://docs.aws.amazon.com/step-functions/latest/dg/amazon-states-language-wait-state.html

            A Wait state (identified by "Type":"Wait") causes the interpreter
            to delay the machine from continuing for a specified time.

            InputPath & OutputPath are allowed (but unusual) in Wait states.
            https://states-language.net/spec.html#statetypes. This is defined
            before on_timeout() so that input is captured in its closure.
            """
            try:
                input = apply_path(data, context, state.get("InputPath", "$"))
            except PathMatchFailure as e:
                handle_error("States.Runtime", str(e))
                self.event_dispatcher.acknowledge(id)
                return

            """
            It's important for this function to be nested as we want the event,
            state and id to be wrapped in its closure, to be used when the
            timeout actually fires.
            """
            def on_timeout():
                try:
                    event["data"] = apply_path(input, context, state.get("OutputPath", "$"))
                    if state.get("End"):
                        handle_terminal_state(state_type, event, id)
                    else:
                        self.change_state(
                            state_machine_type,
                            state_type,
                            state.get("Next"),
                            event
                        )
                        self.event_dispatcher.acknowledge(id)
                except PathMatchFailure as e:
                    handle_error("States.Runtime", str(e))
                    self.event_dispatcher.acknowledge(id)

            """
            Calculate the difference between supplied rfc3339 and current time.
            """
            def get_timeout_from_rfc3339_datetime(rfc3339):
                try:
                    target_timestamp = parse_rfc3339_datetime(rfc3339).timestamp()
                    current_timestamp = time.time()
                    return (target_timestamp - current_timestamp) * 1000
                except ValueError as e:
                    self.logger.warning(
                        "timestamp {} failed to parse correctly, "
                        "defaulting to zero delay".format(rfc3339)
                    )
                    return 0

            """
            The time can be specified as a wait duration, specified in seconds,
            or an absolute expiry time, specified as an ISO-8601 extended offset
            date-time format string.

            The wait duration does not need to be hardcoded and may also be a
            Reference Path to the effective input such as
            "TimestampPath": "$.expirydate"
    
            A Wait state MUST contain exactly one of ”Seconds”, “SecondsPath”,
            “Timestamp”, or “TimestampPath”.
            """
            seconds = state.get("Seconds")
            seconds_path = state.get("SecondsPath")
            timestamp = state.get("Timestamp")
            timestamp_path = state.get("TimestampPath")

            """
            For relative timeouts (seconds and seconds_path) we assume that the
            required delay is relative to the time the state was entered, so we
            add to the entry timestamp converted to epoch time and subtract
            current epoch time. For normal operation this should be extremely
            close to just using the seconds value, but for case where there is
            a backlog on the event queue or the state engine has died and the
            event has been redelivered to another instance then the actual delay
            could be somewhat less. The ASL specification says nothing about
            such a scenario, but it kind of seems logical for the wait time of
            a Wait state to be relative to when the state was actually entered.
            """
            entered_time = context["State"].get("EnteredTime")
            state_timestamp = parse_rfc3339_datetime(entered_time).timestamp()

            try:
                timeout = 0
                if seconds:
                    current_timestamp = time.time()
                    timeout = (state_timestamp + seconds - current_timestamp) * 1000
                elif seconds_path:
                    seconds = apply_path(input, context, seconds_path)
                    if seconds:
                        current_timestamp = time.time()
                        timeout = (state_timestamp + seconds - current_timestamp) * 1000
                elif timestamp:
                    timeout = get_timeout_from_rfc3339_datetime(timestamp)
                elif timestamp_path:
                    timestamp = apply_path(input, context, timestamp_path)
                    if timestamp:
                        timeout = get_timeout_from_rfc3339_datetime(timestamp)

                # Negative timeouts could occur if messages are redelivered or
                # backlogged so we set them to 0 if that happens.
                timeout = timeout if timeout > 0 else 0

                """
                Schedule the timeout. This is slightly subtle, the idea is that
                the event instance, state and id for this call are wrapped in
                the on_timeout function's closure, so when the timeout fires the
                correct event should be published and the correct id acknowledged.
                """
                self.event_dispatcher.set_timeout(on_timeout, timeout)
            except PathMatchFailure as e:
                handle_error("States.Runtime", str(e))
                self.event_dispatcher.acknowledge(id)

        def asl_state_Succeed():
            """
            https://states-language.net/spec.html#succeed-state
            https://docs.aws.amazon.com/step-functions/latest/dg/amazon-states-language-succeed-state.html

            The Succeed State (identified by "Type":"Succeed") terminates a state
            machine successfully. The Succeed State is a useful target for Choice-
            state branches that don't do anything but terminate the machine.
    
            Because Succeed States are terminal states, they have no “Next” field.
            
            InputPath & OutputPath are allowed (but unusual) in Succeed states.
            https://states-language.net/spec.html#statetypes
            """
            try:
                input = apply_path(data, context, state.get("InputPath", "$"))
                event["data"] = apply_path(input, context, state.get("OutputPath", "$"))
                handle_terminal_state(state_type, event, id)
            except PathMatchFailure as e:
                handle_error("States.Runtime", str(e))
                self.event_dispatcher.acknowledge(id)

        def asl_state_Fail():
            """
            https://states-language.net/spec.html#fail-state
            https://docs.aws.amazon.com/step-functions/latest/dg/amazon-states-language-fail-state.html

            The Fail State (identified by "Type":"Fail") terminates the machine
            and marks it as a failure.
        
            Because Fail States are terminal states, they have no “Next” field.

            A Fail State MUST have a string field named “Error”, used to provide
            an error name that can be used for error handling (Retry/Catch),
            operational, or diagnostic purposes. A Fail State MUST have a string
            field named “Cause”, used to provide a human-readable message.

            Fail states don't allow InputPath, OutputPath or ResultPath
            """
            event["data"] = {
                "Error": state.get("Error", "Unspecified"),
                "Cause": state.get("Cause", "Unspecified"),
            }

            handle_terminal_state(state_type, event, id)

        def asl_state_Parallel():
            """
            https://states-language.net/spec.html#parallel-state
            https://docs.aws.amazon.com/step-functions/latest/dg/amazon-states-language-parallel-state.html

            A Parallel state causes the interpreter to execute each branch
            starting with the state named in its “StartAt” field, as concurrently
            as possible.

            The "wait until each branch terminates (reaches a terminal state)
            before processing the Parallel state's “Next” field." part of the 
            Parallel state is implemented in asl_state_collect_results()

            The Parallel state passes its input (potentially as filtered by the
            “InputPath” field) as the input to each branch’s “StartAt” state.
            """
            try:
                input = apply_path(data, context, state.get("InputPath", "$"))

                """
                https://states-language.net/spec.html#using-paths

                If the “Parameters” field is provided, its value, after the
                extraction and embedding, becomes the effective input.
                """
                parameters = evaluate_payload_template(
                    input, context, state.get("Parameters")
                )

                """
                A Parallel State MUST contain a field named “Branches” which
                is an array whose elements MUST be objects. Each object MUST
                contain fields named “States” and “StartAt” whose meanings are
                exactly like those in the top level of a State Machine.

                Iterate through the branches and launch each branch State Machine.
    
                The index of each branch and the Parallel state's raw input is
                needed to create the final result, so we have to store those
                in the context of the events traversing the branch state
                machines. Moreover, it is possible that we could have Parallel
                or Map states nested in the branch state machines. To cater for
                this we add a "Branch" field to the "State" object in the
                context. The "Branch" field is a list of objects holding Input
                and Index as necessary. Each nested Map or Parallel state can
                append to that list, which behaves like a stack.
                """
                context_state = context["State"]
                if not "Branch" in context_state:
                    context_state["Branch"] = []
                context_state["Branch"].append({})

                branches = state.get("Branches", [])
                for index, branch in enumerate(branches):
                    event["data"] = parameters
                    branch_info = {
                        "Parent": current_state,
                        "Input": data,  # Save the raw input to the Parallel state
                        "Index": index, # Index of the current branch
                    }

                    context_state["Name"] = branch.get("StartAt")
                    context_state["EnteredTime"] = (
                        datetime.now(timezone.utc).astimezone().isoformat()
                    )
                    context_state["Branch"][-1] = branch_info

                    self.event_dispatcher.publish(event)

                self.event_dispatcher.acknowledge(id)
            except IntrinsicFailure as e:
                handle_error("States.IntrinsicFailure", str(e))
                self.event_dispatcher.acknowledge(id)
            except PathMatchFailure as e:
                handle_error("States.Runtime", str(e))
                self.event_dispatcher.acknowledge(id)

        def get_start_index(context):
            """
            Boilerplate to retrieve the "Start" index of the Map Iterator. This
            field is used in the implementation of MaxConcurrency. The idea is
            when we process the Map state we launch MaxConcurrency child state
            machine executions and when those complete we re-enter the Map state
            and process the next block of MaxConcurrency, and so on. The "Start"
            field, if present, holds the index of the start of the next block
            to be processed. The "Branch" metadata is a list so we can handle
            the case of nested Map and Parallel states.
            """
            start = 0
            context_state = context["State"]
            if "Branch" in context_state and len(context_state["Branch"]):
                start = context_state["Branch"][-1].get("Start", 0)
            return start

        def asl_state_Map():
            """
            https://states-language.net/spec.html#map-state
            https://docs.aws.amazon.com/step-functions/latest/dg/amazon-states-language-map-state.html

            A Map State causes the interpreter to process all the elements of an
            array, potentially in parallel, with the processing of each element
            independent of the others. The specification uses the term “iteration”
            to describe each such nested execution.

            The Parallel state applies multiple different state-machine branches
            to the same input, while the Map state applies a single state machine
            to multiple input elements.

            The “InputPath” field operates as usual, selecting part of the raw input .
            """
            try:
                input = apply_path(data, context, state.get("InputPath", "$"))

                """
                The “ItemsPath” field’s value is a reference path identifying
                where in the effective input the array field is found.

                The default value of “ItemsPath” is “$”, which is to say the
                whole effective input. So, if a Map State has neither an
                “InputPath” nor a “ItemsPath” field, it is assuming that the
                raw input to the state will be a JSON array.
                """
                items_path = apply_path(input, context, state.get("ItemsPath", "$"))
                if not isinstance(items_path, list):
                    items_path = []

                length = len(items_path)

                """
                A Map State MUST contain an object field named “Iterator” which
                MUST contain fields named “States” and “StartAt”, whose meanings
                are exactly like those in the top level of a State Machine.

                The “Iterator” field’s value is an object that defines a state
                machine which will process each element of the array.
                """
                iterator = state.get("Iterator", {})

                """
                The “MaxConcurrency” field’s value is an integer that provides
                an upper bound on how many invocations of the Iterator may run
                in parallel.
                """
                max_concurrency = state.get("MaxConcurrency", 0)
                if max_concurrency == 0:
                    max_concurrency = length

                """
                Save these values from the Map state. We want to use these
                values in the context evaluated by the parameter evaluation.
                """
                context_state = context["State"]
                map_state_name = context_state["Name"]
                map_state_entered = context_state["EnteredTime"]

                """
                Iterate through the items_path and launch each item's State
                Machine. The index and value of each item is needed to evaluate
                the parameter, so we store it in the "Map" field of the context
                as described in the AWS documentation for the context object.
                https://docs.aws.amazon.com/step-functions/latest/dg/input-output-contextobject.html

                Also, like the Parallel state, the index of each branch and the
                Map state's raw input is needed to create the final result, so
                we again add a "Branch" field to the "State" object in the context.

                The start variable is used with MaxConcurrency. We collect
                results for each "block" of MaxConcurrency results and when all
                have been returned we publish a new event to re-enter the Map
                state, with a start incremented by MaxConcurrency. We carry on
                collecting blocks of MaxConcurrency and re-entering the Map
                state with updated start until we have finished processing
                all items in items_path.
                """
                if length and not "Branch" in context_state:
                    context_state["Branch"] = []

                start = get_start_index(context)
                if length and start == 0:
                    context_state["Branch"].append({})

                end = min(start + max_concurrency, length)

                """
                Need to iterate through the items_path list between start and
                end indices and also retrieve the current index. This is quite
                fiddly in Python when compared to doing similar in lower level
                languages. 
                https://stackoverflow.com/questions/34384523/how-can-loop-through-a-list-from-a-certain-index
                """
                for index, item in enumerate(items_path[start:end], start=start):
                    """
                    Ensure the context evaluated by parameter evaluation is that
                    of the parent Map state not the new Iterator StartAt state.
                    """
                    context_state["Name"] = map_state_name
                    context_state["EnteredTime"] = map_state_entered

                    params = state.get("Parameters")
                    if params:
                        # Store the index and value in the context as described above.
                        context["Map"] = {
                            "Item": {
                                "Index": index,
                                "Value": item,
                            },
                        }

                        """
                        https://states-language.net/spec.html#using-paths

                        If the “Parameters” field is provided, its value, after
                        extraction and embedding, becomes the effective input.
                        """
                        parameters = evaluate_payload_template(
                            input, context, params
                        )

                        del context["Map"]  # Delete after parameters have been processed
                    else:
                        """
                        If no parameters are supplied the effective input to the
                        iteration is the current item i.e $$.Map.Item.Value
                        """
                        parameters = item

                    event["data"] = parameters
                    branch_info = {
                        "Parent": current_state,
                        "Input": data,     # Save the raw input to the Map state
                        "Index": index,    # Index of the current branch
                        "Length": length,  # Number of branches/Iterator instances
                    }
                    if start:
                        branch_info["Start"] = start

                    context_state["Name"] = iterator.get("StartAt")
                    context_state["EnteredTime"] = (
                        datetime.now(timezone.utc).astimezone().isoformat()
                    )
                    context_state["Branch"][-1] = branch_info

                    self.event_dispatcher.publish(event)

                """
                It's not clear what the correct course of action is when the
                effective input is empty, none of the documentation covers this.
                Most likely the result is just an empty array, need to check...
                """
                if length:
                    self.event_dispatcher.acknowledge(id)
                else:
                    """
                    https://states-language.net/spec.html#using-paths

                    The value of "ResultSelector" MUST be a Payload Template,
                    whose input is the result, and whose payload replaces and
                    becomes the effective result.
                    """
                    result = evaluate_payload_template(
                        [], context, state.get("ResultSelector")
                    )

                    # Parallel and Map states apply ResultPath to "raw input"
                    event["data"] = merge_result(data, context, result, state)

                    if state.get("End"):
                        handle_terminal_state(state_type, event, id)
                    else:
                        self.change_state(
                            state_machine_type,
                            state_type,
                            state.get("Next"),
                            event
                        )
                        self.event_dispatcher.acknowledge(id)
            except IntrinsicFailure as e:
                handle_error("States.IntrinsicFailure", str(e))
                self.event_dispatcher.acknowledge(id)
            except PathMatchFailure as e:
                handle_error("States.Runtime", str(e))
                self.event_dispatcher.acknowledge(id)
            except ResultPathMatchFailure as e:
                handle_error("States.ResultPathMatchFailure", str(e))
                self.event_dispatcher.acknowledge(id)

        def asl_state_collect_results(state_type):
            """
            Collect the results from the branches of Parallel and Map states.
            Wait until every branch terminates (reaches a terminal state) before
            processing the Parallel or Map state's “Next” field.
            """
            # Get data object from event again to ensure we have result not input.
            data = event["data"]

            # Retrieve the index, input, length, etc. info for the current branch.
            branch_info = context["State"]["Branch"][-1]  # Top of stack (last item)
            index = branch_info["Index"]

            current_state = branch_info["Parent"]  # Get parent info from "stack"

            state, current_state_machine = find_state(ASL["States"], current_state)
            if state == None:  # state should be valid by this point
                self.log_and_drop("State {} does not exist", current_state, id)
                return

            previous_state_type = state_type
            state_type = state["Type"]

            execution_arn = context["Execution"]["Id"]
            # Initialise branch_results if necessary
            if not execution_arn in self.branch_results:
                self.branch_results[execution_arn] = {}
            if not current_state in self.branch_results[execution_arn]:
                if "Branches" in state:  # Parallel
                    length = len(state["Branches"])
                else:  # Map
                    length = branch_info["Length"]

                self.branch_results[execution_arn][current_state] = {
                    "results": [None]*length,
                    "ids": [None]*length,
                }

            """
            Record the event id so we can acknowledge the event when the
            results from all branches have eventually been returned.
            """
            branch_results = self.branch_results[execution_arn][current_state]
            result = branch_results["results"]
            event_ids = branch_results["ids"]
            result[index] = data
            if previous_state_type != "Parallel" and previous_state_type != "Map":
                event_ids[index] = id

            #print("asl_state_collect_results")
            #print(result)
            
            #print("----------")
            #print(context["State"]["Branch"])
            #print("----------")

            """
            If we haven't yet received the results from all branches check if
            MaxConcurrency has been set and if it has check if we've received
            all of the results for the current block of MaxConcurrency. If we
            haven't then simply return, but if we have then create and publish
            an event to re-enter the Map state with an updated "Start" index
            in order to process the next block of MaxConcurrency.
            """
            if None in result:
                max_concurrency = state.get("MaxConcurrency", 0)
                if max_concurrency:
                    start = branch_info.get("Start", 0)
                    end = min(start + max_concurrency, len(result))

                    partial = result[start:end]
                    if not None in partial:
                        """
                        If we've got all results for a batch of max_concurrency
                        send an event to re-enter the Map state and trigger
                        processing of the next batch.
                        """
                        event["data"] = branch_info["Input"]  # Get saved raw input
                        context["State"]["Name"] = current_state
                        context["State"]["Branch"][-1] = {"Start": end}

                        self.event_dispatcher.publish(event)

                return

            """
            Remove last item from the "Branch" list, then if it becomes empty
            delete "Branch" from context as we've finished with it.
            """
            del context["State"]["Branch"][-1]
            if len(context["State"]["Branch"]) == 0:
                del context["State"]["Branch"]

            context["State"]["Name"] = current_state

            # Parallel and Map states apply ResultPath to "raw input"
            data = branch_info["Input"]  # Get saved raw input

            """
            https://states-language.net/spec.html#using-paths

            The value of "ResultSelector" MUST be a Payload Template, whose
            input is the result, and whose payload replaces and becomes the
            effective result.
            """
            try:
                result = evaluate_payload_template(
                    result, context, state.get("ResultSelector")
                )

                event["data"] = merge_result(data, context, result, state)
            except IntrinsicFailure as e:
                handle_error("States.IntrinsicFailure", str(e))
                # Acknowledge the events for each branch's terminal state
                for event_id in event_ids:
                    if event_id:
                        self.event_dispatcher.acknowledge(event_id)
                return
            except PathMatchFailure as e:
                handle_error("States.Runtime", str(e))
                # Acknowledge the events for each branch's terminal state
                for event_id in event_ids:
                    if event_id:
                        self.event_dispatcher.acknowledge(event_id)
                return
            except ResultPathMatchFailure as e:
                handle_error("States.ResultPathMatchFailure", str(e))
                # Acknowledge the events for each branch's terminal state
                for event_id in event_ids:
                    if event_id:
                        self.event_dispatcher.acknowledge(event_id)
                return

            """
            Publish any new state change before acknowledging the events.
            """
            if not state.get("End"):
                self.change_state(
                    state_machine_type,
                    state_type,
                    state.get("Next"),
                    event
                )

            # Acknowledge the events for each branch's terminal state
            for event_id in event_ids:
                if event_id:
                    self.event_dispatcher.acknowledge(event_id)

            """
            Need to do this *after* acknowledging the events as it deletes the
            Parallel or Map branch results for the current execution.
            """
            if state.get("End"):
                handle_terminal_state(state_type, event)


        """
        End of nested state handler functions.
        """
        # ----------------------------------------------------------------------


        """
        Update the execution history with StateEntered information. The test
        for RetryCount is because when we trigger a retry we do so by publishing
        an event to the message queue that triggers a transition back to the
        state being retried, but as this represents a *retry* and not an actual
        entry to the state we want to suppress the history update.
        Similarly, the test for reentered_map is because if MaxConcurrency is
        set we will re-enter the Map state, possibly several times, to process
        the next batch of items so again we want to suppress the history update.
        """
        execution_arn = context["Execution"]["Id"]
        reentered_map = state_type == "Map" and get_start_index(context) != 0
        if not context["State"].get("RetryCount") and not reentered_map:
            self.update_execution_history(
                state_machine_type,
                execution_arn,
                state_type + "StateEntered",
                {"input": json.dumps(data), "name": current_state},
            )
        # self.logger=init_logging(log_name="asl_execution_history")
        # self.logger.info(self.execution_history)
        """
        We check the execution history hasn't exceeded the 25000 event limit in
        https://docs.aws.amazon.com/step-functions/latest/dg/limits.html
        here rather than in update_execution_history, because here we can more
        cleanly call handle_error() and fail the execution in a way that is
        consistent with other failures. Note that this approach may result in
        the stored history being a few events more than the actual limit. Doing
        the test here also means that ExecutionSucceeded/ExecutionFailed events
        won't cause the execution history limit to be exceeded, as failing the
        execution on those events doesn't seem particularly sensible.
        The error code of States.ExecutionHistoryLimitExceeded is chosen to
        mirror the code States.DataLimitExceeded used when the input or output
        size has exceeded their limit, however the actual error code used for
        this failure unfortunately doesn't seem to be documented anywhere.
        """
        if state_machine_type == "STANDARD":
            execution_history_length = len(self.execution_history[execution_arn])
            #print(execution_history_length)
            if execution_history_length > MAX_EXECUTION_HISTORY_LENGTH:
                message = ("Execution '{}' has a history of {} events and has "
                           "exceeded the maximum execution history size.").format(
                            execution_arn, execution_history_length
                          )
                self.logger.error(message)
                handle_error("States.ExecutionHistoryLimitExceeded", message)
                self.event_dispatcher.acknowledge(id)
                return

        """
        Use the ASL state type of the current state to dynamically invoke the
        appropriate ASL state handler given state type. The (Python) lambda
        provides a default handler in case of malformed ASL. 
        """
        locals().get(
            "asl_state_" + state_type,
            lambda: self.logger.error(
                "StateEngine illegal state transition: {}".format(state_type)
            ),
        )()




#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
# 
#   http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
# Run with:
# PYTHONPATH=.. python3 producer_consumer_asyncio_topic.py
# PYTHONPATH=.. LOG_LEVEL=DEBUG python3 producer_consumer_asyncio_topic.py
#
"""
This example application is mainly for exercising the messaging API and isn't
directly related to the ASL engine per se. It is an asyncio example that creates
a bounded queue and runs a simple producer/consumer as coroutines.
"""

import sys
assert sys.version_info >= (3, 6) # Bomb out if not running Python3.6

import asyncio, json, time

from asl_workflow_engine.logger import init_logging
from asl_workflow_engine.amqp_0_9_1_messaging_asyncio import Connection, Message
from asl_workflow_engine.messaging_exceptions import *

class Worker():

    def __init__(self, name):
        self.name = name
        # Initialise logger
        self.logger = init_logging(log_name=name)
        self.count = 0

    def handler(self, message):
        #print(message)
        #print(message.body)
        self.count += 1
        print(self.count)

        # Safer but slower alternative to connection.session(auto_ack=True)
        #if self.count % 10 == 0:  # Periodically acknowledge consumed messages
        #    message.acknowledge()

    async def run(self):
        # Connect to worker queue and process data.
        connection = Connection("amqp://localhost:5672?connection_attempts=20&retry_delay=10&heartbeat=0")
        try:
            await connection.open()
            #session = await connection.session()
            session = await connection.session(auto_ack=True)   

            """
            This illustrates a fairly "fancy" Address String. At its heart it's
            subscribing to the subject news.* on test_topic but the extended
            syntax is creating an AMQP "node" that represent the exchange which
            allows us to create test_topic if it doesn't already exist and
            whilst it's usual for topic "subscription queues" to be quite
            anonymous with broker generated names, in this case we use the AMQP
            "link" to allow us to specify a named subscription queue with
            additional arguments. It's a little bit contrived, but serves to
            illustrate the syntax.
            """
            self.consumer = await session.consumer(
                "test_topic/news.*" + '; {"node": {"x-declare": {"exchange": "test_topic", "exchange-type": "topic", "auto-delete": true}}, "link": {"x-declare": {"queue": "' + self.name + '", "arguments": {"x-max-length": 10, "x-overflow": "reject-publish"}}}}'
            )

            self.consumer.capacity = 100; # Enable consumer prefetch
            await self.consumer.set_message_listener(self.handler)

            self.producer = await session.producer("test_topic")

            self.producer.enable_exceptions(sync=True)
            #self.producer.enable_exceptions()

            waiters = []
            for i in range(1000000):
                #print("message {}".format(self.producer.next_publish_seq_no))
                message = Message(body="hello message {}".format(
                    self.producer.next_publish_seq_no)
                )
                message.subject = "news.uk"
                #print(message)

                try:
                    # When producer.enable_exceptions(True) is set the send()
                    # method is made awaitable by returning a Future, resolved
                    # when the broker acks the message or exceptioned if the
                    # broker nacks. For fully "synchronous" publishing one can
                    # simply do: await self.producer.send(message)
                    # but that has a serious throughput/latency impact waiting
                    # for the publisher-confirms from the broker. The impact is
                    # especially bad for the case of durable queues.
                    # https://www.rabbitmq.com/confirms.html#publisher-confirms
                    # https://www.rabbitmq.com/confirms.html#publisher-confirms-latency
                    # To mitigate this the example below publishes in batches
                    # (optimally the batch size should be roughly the same as
                    # the session capacity). We store the awaitables in a list
                    # then periodically do an asyncio.gather of the waiters
                    waiters.append(self.producer.send(message))
                    if i % 100 == 0:
                        await asyncio.gather(*waiters)
                        waiters = []

                    # When producer.enable_exceptions(False) is set the send()
                    # method will raise an exception that contains the sequence
                    # numbers of messages that have failed to publish. If no
                    # call to enable_exceptions() is made send() just does
                    # basic_publish() with no confirmation, which could result
                    # in message loss. The asyncio.sleep() in the example is
                    # necessary to periodically yield to allow consumer delivery.
                    # If an exception is thrown either increase the sleep or
                    # catch the exception and resend the nacked messages that
                    # have been recorded in exception.undelivered
                    """
                    self.producer.send(message)
                    if i % 100 == 0:
                        await asyncio.sleep(.006)
                    """
                except SendError as e:
                    #print(waiters)
                    raise e

            await connection.start(); # Wait until connection closes.
    
        except MessagingError as e:  # ConnectionError, SessionError etc.
            self.logger.error(e)

        connection.close();

def global_exception_handler(loop, context):
    # context["message"] will always be there; but context["exception"] may not
    #msg = context.get("exception", context["message"])
    #print("********** global_exception_handler **********")
    #print(msg)
    pass  # Just swallow "exception was never retrieved"

if __name__ == '__main__':
    workers = ["SubscriptionQueue"]
    loop = asyncio.get_event_loop()
    loop.set_exception_handler(global_exception_handler)
    # Create list of Worker tasks using list comprehension
    tasks = [loop.create_task(Worker(name=w).run()) for w in workers]
    # Run the tasks concurrently
    loop.run_until_complete(asyncio.gather(*tasks))
    loop.close()
    



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
# 
#   http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

from setuptools import setup, find_packages

setup(
    name="local_workflow_engine",
    version="1.0.0",
    description="A workflow engine based on the Amazon States Language (ASL).",
    long_description="A workflow engine based on the Amazon States Language (ASL). It is intended to provide the features and API of AWS Step Functions in an engine that can be deployed to a range of different hosting environments.",
    packages=find_packages(),
    install_requires=["pika",
                      "structlog",
                      "ujson",
                      "jsonpath",
                      "flask",
                      "quart",
                      "redis",
                      "pottery",
                      "opentracing>=2.2",
                      "aioprometheus",
                      "jaeger_client"]
)
